# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower message-based FedXgbBagging strategy."""
from collections.abc import Iterable
from typing import cast

import numpy as np

from flwr.common import ArrayRecord, ConfigRecord, Message, MetricRecord
from flwr.server import Grid

from ..exception import InconsistentMessageReplies
from .fedavg import FedAvg
from .strategy_utils import aggregate_bagging


# pylint: disable=line-too-long
class FedXgbBagging(FedAvg):
    """Configurable FedXgbBagging strategy implementation.

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    """

    current_bst: bytes | None = None

    def _ensure_single_array(self, arrays: ArrayRecord) -> None:
        """Check that ensures there's only one Array in the ArrayRecord."""
        n = len(arrays)
        if n != 1:
            raise InconsistentMessageReplies(
                reason="Expected exactly one Array in ArrayRecord. "
                "Skipping aggregation."
            )

    def configure_train(
        self, server_round: int, arrays: ArrayRecord, config: ConfigRecord, grid: Grid
    ) -> Iterable[Message]:
        """Configure the next round of federated training."""
        self._ensure_single_array(arrays)
        # Keep track of array record being communicated
        self.current_bst = arrays["0"].numpy().tobytes()
        return super().configure_train(server_round, arrays, config, grid)

    def aggregate_train(
        self,
        server_round: int,
        replies: Iterable[Message],
    ) -> tuple[ArrayRecord | None, MetricRecord | None]:
        """Aggregate ArrayRecords and MetricRecords in the received Messages."""
        valid_replies, _ = self._check_and_log_replies(replies, is_train=True)

        arrays, metrics = None, None
        if valid_replies:
            reply_contents = [msg.content for msg in valid_replies]
            array_record_key = next(iter(reply_contents[0].array_records.keys()))

            # Aggregate ArrayRecords
            for content in reply_contents:
                self._ensure_single_array(cast(ArrayRecord, content[array_record_key]))
                bst = content[array_record_key]["0"].numpy().tobytes()  # type: ignore[union-attr]

                if self.current_bst is not None:
                    self.current_bst = aggregate_bagging(self.current_bst, bst)

            if self.current_bst is not None:
                arrays = ArrayRecord([np.frombuffer(self.current_bst, dtype=np.uint8)])

            # Aggregate MetricRecords
            metrics = self.train_metrics_aggr_fn(
                reply_contents,
                self.weighted_by_key,
            )
        return arrays, metrics



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Federated Median (FedMedian) [Yin et al., 2018] strategy.

Paper: arxiv.org/pdf/1803.01498v1.pdf
"""


from collections.abc import Iterable
from typing import cast

import numpy as np

from flwr.common import Array, ArrayRecord, Message, MetricRecord

from .fedavg import FedAvg


class FedMedian(FedAvg):
    """Federated Median (FedMedian) strategy.

    Implementation based on https://arxiv.org/pdf/1803.01498v1

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    """

    def aggregate_train(
        self,
        server_round: int,
        replies: Iterable[Message],
    ) -> tuple[ArrayRecord | None, MetricRecord | None]:
        """Aggregate ArrayRecords and MetricRecords in the received Messages."""
        # Call FedAvg aggregate_train to perform validation and aggregation
        valid_replies, _ = self._check_and_log_replies(replies, is_train=True)

        if not valid_replies:
            return None, None

        # Aggregate ArrayRecords using median
        # Get the key for the only ArrayRecord from the first Message
        record_key = list(valid_replies[0].content.array_records.keys())[0]
        # Preserve keys for arrays in ArrayRecord
        array_keys = list(valid_replies[0].content[record_key].keys())

        # Compute median for each layer and construct ArrayRecord
        arrays = ArrayRecord()
        for array_key in array_keys:
            # Get the corresponding layer from each client
            layers = [
                cast(ArrayRecord, msg.content[record_key]).pop(array_key).numpy()
                for msg in valid_replies
            ]
            # Compute median and save as Array in ArrayRecord
            arrays[array_key] = Array(np.median(np.stack(layers), axis=0))

        # Aggregate MetricRecords
        metrics = self.train_metrics_aggr_fn(
            [msg.content for msg in valid_replies],
            self.weighted_by_key,
        )
        return arrays, metrics



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface `federation show` command."""


import io
from pathlib import Path
from typing import Annotated, Any

import typer
from rich.console import Console
from rich.table import Table
from rich.text import Text

from flwr.cli.config_utils import (
    exit_if_no_address,
    load_and_validate,
    process_loaded_project_config,
    validate_federation_in_project_config,
)
from flwr.cli.ls import _get_status_style
from flwr.common.constant import FAB_CONFIG_FILE, NOOP_ACCOUNT_NAME, CliOutputFormat
from flwr.common.logger import print_json_error, redirect_output, restore_output
from flwr.common.serde import run_from_proto
from flwr.proto.control_pb2 import (  # pylint: disable=E0611
    ShowFederationRequest,
    ShowFederationResponse,
)
from flwr.proto.control_pb2_grpc import ControlStub
from flwr.proto.node_pb2 import NodeInfo  # pylint: disable=E0611
from flwr.supercore.constant import NOOP_FEDERATION

from ..run_utils import RunRow, format_runs
from ..utils import flwr_cli_grpc_exc_handler, init_channel, load_cli_auth_plugin


def show(  # pylint: disable=R0914, R0913, R0917
    app: Annotated[
        Path,
        typer.Argument(help="Path of the Flower project"),
    ] = Path("."),
    federation: Annotated[
        str | None,
        typer.Argument(help="Name of the federation"),
    ] = None,
    output_format: Annotated[
        str,
        typer.Option(
            "--format",
            case_sensitive=False,
            help="Format output using 'default' view or 'json'",
        ),
    ] = CliOutputFormat.DEFAULT,
) -> None:
    """Show details of a federation.

    Display comprehensive information about a federation including its members,
    registered SuperNodes, and runs.
    """
    suppress_output = output_format == CliOutputFormat.JSON
    captured_output = io.StringIO()
    try:
        if suppress_output:
            redirect_output(captured_output)
        typer.secho("Loading project configuration... ", fg=typer.colors.BLUE)

        pyproject_path = app / FAB_CONFIG_FILE if app else None
        config, errors, warnings = load_and_validate(pyproject_path, check_module=False)
        config = process_loaded_project_config(config, errors, warnings)
        federation, federation_config = validate_federation_in_project_config(
            federation, config
        )
        exit_if_no_address(federation_config, "federation show")
        real_federation: str = federation_config.get("federation", NOOP_FEDERATION)
        channel = None
        try:
            auth_plugin = load_cli_auth_plugin(app, federation, federation_config)
            channel = init_channel(app, federation_config, auth_plugin)
            stub = ControlStub(channel)
            typer.echo(f"📄 Showing '{real_federation}' federation ...")
            members, nodes, runs = _show_federation(stub, real_federation)
            restore_output()
            if output_format == CliOutputFormat.JSON:
                Console().print_json(data=_to_json(members, nodes, runs))
            else:
                Console().print(_to_members_table(members))
                Console().print(_to_nodes_table(nodes))
                Console().print(_to_runs_table(runs))
        finally:
            if channel:
                channel.close()
    except (typer.Exit, Exception) as err:  # pylint: disable=broad-except
        if suppress_output:
            restore_output()
            e_message = captured_output.getvalue()
            print_json_error(e_message, err)
        else:
            typer.secho(
                f"{err}",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
    finally:
        if suppress_output:
            restore_output()
        captured_output.close()


def _show_federation(
    stub: ControlStub, federation: str
) -> tuple[list[str], list[NodeInfo], list[RunRow]]:
    """Show federation details.

    Parameters
    ----------
    stub : ControlStub
        The gRPC stub for Control API communication.
    federation : str
        Name of the federation to show.

    Returns
    -------
    tuple[list[str], list[NodeInfo], list[RunRow]]
        A tuple containing (member_account_ids, nodes, runs).
    """
    with flwr_cli_grpc_exc_handler():
        res: ShowFederationResponse = stub.ShowFederation(
            ShowFederationRequest(federation_name=federation)
        )

    fed_proto = res.federation
    runs = [run_from_proto(run_proto) for run_proto in fed_proto.runs]
    formatted_runs = format_runs(runs, res.now)

    return list(fed_proto.member_aids), list(fed_proto.nodes), formatted_runs


def _to_members_table(member_aids: list[str]) -> Table:
    """Format the provided list of federation members as a rich Table.

    Parameters
    ----------
    member_aids : list[str]
        List of member account identifiers.

    Returns
    -------
    Table
        Rich Table object with formatted member information.
    """
    table = Table(title="Federation Members", header_style="bold cyan", show_lines=True)

    table.add_column(
        Text("Account ID", justify="center"), style="bright_black", no_wrap=True
    )
    table.add_column(Text("Role", justify="center"), style="bright_black", no_wrap=True)

    for member_aid in member_aids:
        table.add_row(member_aid, "Member")

    return table


def _to_nodes_table(nodes: list[NodeInfo]) -> Table:
    """Format the provided list of federation nodes as a rich Table.

    Parameters
    ----------
    nodes : list[NodeInfo]
        List of NodeInfo objects containing node details.

    Returns
    -------
    Table
        Rich Table object with formatted node information.

    Raises
    ------
    ValueError
        If an unexpected node status is encountered.
    """
    table = Table(
        title="SuperNodes in the Federation", header_style="bold cyan", show_lines=True
    )

    # Add columns
    table.add_column(
        Text("Node ID", justify="center"), style="bright_black", no_wrap=True
    )
    table.add_column(Text("Owner", justify="center"))
    table.add_column(Text("Status", justify="center"))

    for row in nodes:
        owner_name = row.owner_name
        status = row.status

        if status == "online":
            status_style = "green"
        elif status == "offline":
            status_style = "bright_yellow"
        elif status == "unregistered":
            continue
        elif status == "registered":
            status_style = "blue"
        else:
            raise ValueError(f"Unexpected node status '{status}'")

        formatted_row = (
            f"[bold]{row.node_id}[/bold]",
            (
                f"{owner_name}"
                if owner_name != NOOP_ACCOUNT_NAME
                else f"[dim]{owner_name}[/dim]"
            ),
            f"[{status_style}]{status}",
        )
        table.add_row(*formatted_row)

    return table


def _to_runs_table(run_list: list[RunRow]) -> Table:
    """Format the provided list of federation runs as a rich Table.

    Parameters
    ----------
    run_list : list[RunRow]
        List of RunRow objects containing run details.

    Returns
    -------
    Table
        Rich Table object with formatted run information.
    """
    table = Table(
        title="Runs in the Federation", header_style="bold cyan", show_lines=True
    )

    # Add columns
    table.add_column(Text("Run ID", justify="center"), no_wrap=True)
    table.add_column(Text("App", justify="center"))
    table.add_column(Text("Status", justify="center"))
    table.add_column(Text("Elapsed", justify="center"), style="blue")

    for row in run_list:
        status_style = _get_status_style(row.status_text)

        formatted_row = (
            f"[bold]{row.run_id}[/bold]",
            f"@{row.fab_id}=={row.fab_version}",
            f"[{status_style}]{row.status_text}[/{status_style}]",
            row.elapsed,
        )
        table.add_row(*formatted_row)

    return table


def _to_json(
    members: list[str], nodes: list[NodeInfo], runs: list[RunRow]
) -> list[list[dict[str, str]]]:
    """Format the provided federation information to JSON serializable format.

    Parameters
    ----------
    members : list[str]
        List of member account identifiers.
    nodes : list[NodeInfo]
        List of NodeInfo objects.
    runs : list[RunRow]
        List of RunRow objects.

    Returns
    -------
    list[list[dict[str, str]]]
        Nested list containing dictionaries for members, nodes, and runs.
    """
    members_list: list[dict[str, Any]] = []
    nodes_list: list[dict[str, Any]] = []
    runs_list: list[dict[str, Any]] = []

    for member in members:
        members_list.append({"member_id": member, "role": "Member"})

    for node in nodes:
        nodes_list.append(
            {
                "node_id": node.node_id,
                "owner": node.owner_name,
                "status": node.status,
            }
        )

    for run in runs:
        runs_list.append(
            {
                "run_id": run.run_id,
                "app": f"@{run.fab_id}=={run.fab_version}",
                "status": run.status_text,
                "elapsed": run.elapsed,
            }
        )

    return [members_list, nodes_list, runs_list]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2022 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import abc
import collections.abc
import flwr.proto.fab_pb2
import flwr.proto.fleet_pb2
import flwr.proto.heartbeat_pb2
import flwr.proto.message_pb2
import flwr.proto.run_pb2
import grpc
import grpc.aio
import typing

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta): ...

class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

class FleetStub:
    def __init__(self, channel: typing.Union[grpc.Channel, grpc.aio.Channel]) -> None: ...
    RegisterNode: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.RegisterNodeFleetRequest,
        flwr.proto.fleet_pb2.RegisterNodeFleetResponse,
    ]
    """Register Node"""

    ActivateNode: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.ActivateNodeRequest,
        flwr.proto.fleet_pb2.ActivateNodeResponse,
    ]
    """Activate Node"""

    DeactivateNode: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.DeactivateNodeRequest,
        flwr.proto.fleet_pb2.DeactivateNodeResponse,
    ]
    """Deactivate Node"""

    UnregisterNode: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.UnregisterNodeFleetRequest,
        flwr.proto.fleet_pb2.UnregisterNodeFleetResponse,
    ]
    """Unregister Node"""

    SendNodeHeartbeat: grpc.UnaryUnaryMultiCallable[
        flwr.proto.heartbeat_pb2.SendNodeHeartbeatRequest,
        flwr.proto.heartbeat_pb2.SendNodeHeartbeatResponse,
    ]

    PullMessages: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.PullMessagesRequest,
        flwr.proto.fleet_pb2.PullMessagesResponse,
    ]
    """Retrieve one or more messages, if possible

    HTTP API path: /api/v1/fleet/pull-messages
    """

    PushMessages: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.PushMessagesRequest,
        flwr.proto.fleet_pb2.PushMessagesResponse,
    ]
    """Complete one or more messages, if possible

    HTTP API path: /api/v1/fleet/push-messages
    """

    GetRun: grpc.UnaryUnaryMultiCallable[
        flwr.proto.run_pb2.GetRunRequest,
        flwr.proto.run_pb2.GetRunResponse,
    ]

    GetFab: grpc.UnaryUnaryMultiCallable[
        flwr.proto.fab_pb2.GetFabRequest,
        flwr.proto.fab_pb2.GetFabResponse,
    ]
    """Get FAB"""

    PushObject: grpc.UnaryUnaryMultiCallable[
        flwr.proto.message_pb2.PushObjectRequest,
        flwr.proto.message_pb2.PushObjectResponse,
    ]
    """Push Object"""

    PullObject: grpc.UnaryUnaryMultiCallable[
        flwr.proto.message_pb2.PullObjectRequest,
        flwr.proto.message_pb2.PullObjectResponse,
    ]
    """Pull Object"""

    ConfirmMessageReceived: grpc.UnaryUnaryMultiCallable[
        flwr.proto.message_pb2.ConfirmMessageReceivedRequest,
        flwr.proto.message_pb2.ConfirmMessageReceivedResponse,
    ]
    """Confirm Message Received"""

class FleetAsyncStub:
    RegisterNode: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.RegisterNodeFleetRequest,
        flwr.proto.fleet_pb2.RegisterNodeFleetResponse,
    ]
    """Register Node"""

    ActivateNode: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.ActivateNodeRequest,
        flwr.proto.fleet_pb2.ActivateNodeResponse,
    ]
    """Activate Node"""

    DeactivateNode: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.DeactivateNodeRequest,
        flwr.proto.fleet_pb2.DeactivateNodeResponse,
    ]
    """Deactivate Node"""

    UnregisterNode: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.UnregisterNodeFleetRequest,
        flwr.proto.fleet_pb2.UnregisterNodeFleetResponse,
    ]
    """Unregister Node"""

    SendNodeHeartbeat: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.heartbeat_pb2.SendNodeHeartbeatRequest,
        flwr.proto.heartbeat_pb2.SendNodeHeartbeatResponse,
    ]

    PullMessages: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.PullMessagesRequest,
        flwr.proto.fleet_pb2.PullMessagesResponse,
    ]
    """Retrieve one or more messages, if possible

    HTTP API path: /api/v1/fleet/pull-messages
    """

    PushMessages: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fleet_pb2.PushMessagesRequest,
        flwr.proto.fleet_pb2.PushMessagesResponse,
    ]
    """Complete one or more messages, if possible

    HTTP API path: /api/v1/fleet/push-messages
    """

    GetRun: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.run_pb2.GetRunRequest,
        flwr.proto.run_pb2.GetRunResponse,
    ]

    GetFab: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.fab_pb2.GetFabRequest,
        flwr.proto.fab_pb2.GetFabResponse,
    ]
    """Get FAB"""

    PushObject: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.message_pb2.PushObjectRequest,
        flwr.proto.message_pb2.PushObjectResponse,
    ]
    """Push Object"""

    PullObject: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.message_pb2.PullObjectRequest,
        flwr.proto.message_pb2.PullObjectResponse,
    ]
    """Pull Object"""

    ConfirmMessageReceived: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.message_pb2.ConfirmMessageReceivedRequest,
        flwr.proto.message_pb2.ConfirmMessageReceivedResponse,
    ]
    """Confirm Message Received"""

class FleetServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def RegisterNode(
        self,
        request: flwr.proto.fleet_pb2.RegisterNodeFleetRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fleet_pb2.RegisterNodeFleetResponse, collections.abc.Awaitable[flwr.proto.fleet_pb2.RegisterNodeFleetResponse]]:
        """Register Node"""

    @abc.abstractmethod
    def ActivateNode(
        self,
        request: flwr.proto.fleet_pb2.ActivateNodeRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fleet_pb2.ActivateNodeResponse, collections.abc.Awaitable[flwr.proto.fleet_pb2.ActivateNodeResponse]]:
        """Activate Node"""

    @abc.abstractmethod
    def DeactivateNode(
        self,
        request: flwr.proto.fleet_pb2.DeactivateNodeRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fleet_pb2.DeactivateNodeResponse, collections.abc.Awaitable[flwr.proto.fleet_pb2.DeactivateNodeResponse]]:
        """Deactivate Node"""

    @abc.abstractmethod
    def UnregisterNode(
        self,
        request: flwr.proto.fleet_pb2.UnregisterNodeFleetRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fleet_pb2.UnregisterNodeFleetResponse, collections.abc.Awaitable[flwr.proto.fleet_pb2.UnregisterNodeFleetResponse]]:
        """Unregister Node"""

    @abc.abstractmethod
    def SendNodeHeartbeat(
        self,
        request: flwr.proto.heartbeat_pb2.SendNodeHeartbeatRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.heartbeat_pb2.SendNodeHeartbeatResponse, collections.abc.Awaitable[flwr.proto.heartbeat_pb2.SendNodeHeartbeatResponse]]: ...

    @abc.abstractmethod
    def PullMessages(
        self,
        request: flwr.proto.fleet_pb2.PullMessagesRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fleet_pb2.PullMessagesResponse, collections.abc.Awaitable[flwr.proto.fleet_pb2.PullMessagesResponse]]:
        """Retrieve one or more messages, if possible

        HTTP API path: /api/v1/fleet/pull-messages
        """

    @abc.abstractmethod
    def PushMessages(
        self,
        request: flwr.proto.fleet_pb2.PushMessagesRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fleet_pb2.PushMessagesResponse, collections.abc.Awaitable[flwr.proto.fleet_pb2.PushMessagesResponse]]:
        """Complete one or more messages, if possible

        HTTP API path: /api/v1/fleet/push-messages
        """

    @abc.abstractmethod
    def GetRun(
        self,
        request: flwr.proto.run_pb2.GetRunRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.run_pb2.GetRunResponse, collections.abc.Awaitable[flwr.proto.run_pb2.GetRunResponse]]: ...

    @abc.abstractmethod
    def GetFab(
        self,
        request: flwr.proto.fab_pb2.GetFabRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.fab_pb2.GetFabResponse, collections.abc.Awaitable[flwr.proto.fab_pb2.GetFabResponse]]:
        """Get FAB"""

    @abc.abstractmethod
    def PushObject(
        self,
        request: flwr.proto.message_pb2.PushObjectRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.message_pb2.PushObjectResponse, collections.abc.Awaitable[flwr.proto.message_pb2.PushObjectResponse]]:
        """Push Object"""

    @abc.abstractmethod
    def PullObject(
        self,
        request: flwr.proto.message_pb2.PullObjectRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.message_pb2.PullObjectResponse, collections.abc.Awaitable[flwr.proto.message_pb2.PullObjectResponse]]:
        """Pull Object"""

    @abc.abstractmethod
    def ConfirmMessageReceived(
        self,
        request: flwr.proto.message_pb2.ConfirmMessageReceivedRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.message_pb2.ConfirmMessageReceivedResponse, collections.abc.Awaitable[flwr.proto.message_pb2.ConfirmMessageReceivedResponse]]:
        """Confirm Message Received"""

def add_FleetServicer_to_server(servicer: FleetServicer, server: typing.Union[grpc.Server, grpc.aio.Server]) -> None: ...



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Federated Averaging with Momentum (FedAvgM) [Hsu et al., 2019] strategy.

Paper: arxiv.org/pdf/1909.06335.pdf
"""


from collections.abc import Callable
from logging import WARNING

from flwr.common import (
    FitRes,
    MetricsAggregationFn,
    NDArrays,
    Parameters,
    Scalar,
    ndarrays_to_parameters,
    parameters_to_ndarrays,
)
from flwr.common.logger import log
from flwr.server.client_manager import ClientManager
from flwr.server.client_proxy import ClientProxy

from .aggregate import aggregate
from .fedavg import FedAvg


# pylint: disable=line-too-long
class FedAvgM(FedAvg):
    """Federated Averaging with Momentum strategy.

    Implementation based on https://arxiv.org/abs/1909.06335

    Parameters
    ----------
    fraction_fit : float, optional
        Fraction of clients used during training. Defaults to 1.0.
    fraction_evaluate : float, optional
        Fraction of clients used during validation. Defaults to 1.0.
    min_fit_clients : int, optional
        Minimum number of clients used during training. Defaults to 2.
    min_evaluate_clients : int, optional
        Minimum number of clients used during validation. Defaults to 2.
    min_available_clients : int, optional
        Minimum number of total clients in the system. Defaults to 2.
    evaluate_fn : Optional[Callable[[int, NDArrays, Dict[str, Scalar]], Optional[Tuple[float, Dict[str, Scalar]]]]]
        Optional function used for validation. Defaults to None.
    on_fit_config_fn : Callable[[int], Dict[str, Scalar]], optional
        Function used to configure training. Defaults to None.
    on_evaluate_config_fn : Callable[[int], Dict[str, Scalar]], optional
        Function used to configure validation. Defaults to None.
    accept_failures : bool, optional
        Whether or not accept rounds containing failures. Defaults to True.
    initial_parameters : Parameters, optional
        Initial global model parameters.
    server_learning_rate: float
        Server-side learning rate used in server-side optimization.
        Defaults to 1.0.
    server_momentum: float
        Server-side momentum factor used for FedAvgM. Defaults to 0.0.
    """

    # pylint: disable=too-many-arguments,too-many-instance-attributes, line-too-long
    def __init__(
        self,
        *,
        fraction_fit: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_fit_clients: int = 2,
        min_evaluate_clients: int = 2,
        min_available_clients: int = 2,
        evaluate_fn: (
            Callable[
                [int, NDArrays, dict[str, Scalar]],
                tuple[float, dict[str, Scalar]] | None,
            ]
            | None
        ) = None,
        on_fit_config_fn: Callable[[int], dict[str, Scalar]] | None = None,
        on_evaluate_config_fn: Callable[[int], dict[str, Scalar]] | None = None,
        accept_failures: bool = True,
        initial_parameters: Parameters | None = None,
        fit_metrics_aggregation_fn: MetricsAggregationFn | None = None,
        evaluate_metrics_aggregation_fn: MetricsAggregationFn | None = None,
        server_learning_rate: float = 1.0,
        server_momentum: float = 0.0,
    ) -> None:
        super().__init__(
            fraction_fit=fraction_fit,
            fraction_evaluate=fraction_evaluate,
            min_fit_clients=min_fit_clients,
            min_evaluate_clients=min_evaluate_clients,
            min_available_clients=min_available_clients,
            evaluate_fn=evaluate_fn,
            on_fit_config_fn=on_fit_config_fn,
            on_evaluate_config_fn=on_evaluate_config_fn,
            accept_failures=accept_failures,
            initial_parameters=initial_parameters,
            fit_metrics_aggregation_fn=fit_metrics_aggregation_fn,
            evaluate_metrics_aggregation_fn=evaluate_metrics_aggregation_fn,
        )
        self.server_learning_rate = server_learning_rate
        self.server_momentum = server_momentum
        self.server_opt: bool = (self.server_momentum != 0.0) or (
            self.server_learning_rate != 1.0
        )
        self.momentum_vector: NDArrays | None = None

    def __repr__(self) -> str:
        """Compute a string representation of the strategy."""
        rep = f"FedAvgM(accept_failures={self.accept_failures})"
        return rep

    def initialize_parameters(self, client_manager: ClientManager) -> Parameters | None:
        """Initialize global model parameters."""
        return self.initial_parameters

    def aggregate_fit(
        self,
        server_round: int,
        results: list[tuple[ClientProxy, FitRes]],
        failures: list[tuple[ClientProxy, FitRes] | BaseException],
    ) -> tuple[Parameters | None, dict[str, Scalar]]:
        """Aggregate fit results using weighted average."""
        if not results:
            return None, {}
        # Do not aggregate if there are failures and failures are not accepted
        if not self.accept_failures and failures:
            return None, {}
        # Convert results
        weights_results = [
            (parameters_to_ndarrays(fit_res.parameters), fit_res.num_examples)
            for _, fit_res in results
        ]

        fedavg_result = aggregate(weights_results)
        # following convention described in
        # https://pytorch.org/docs/stable/generated/torch.optim.SGD.html
        if self.server_opt:
            # You need to initialize the model
            assert (
                self.initial_parameters is not None
            ), "When using server-side optimization, model needs to be initialized."
            initial_weights = parameters_to_ndarrays(self.initial_parameters)

            # remember that updates are the opposite of gradients
            pseudo_gradient: NDArrays = [
                x - y
                for x, y in zip(
                    parameters_to_ndarrays(self.initial_parameters),
                    fedavg_result,
                    strict=True,
                )
            ]
            if self.server_momentum > 0.0:
                if server_round > 1:
                    assert (
                        self.momentum_vector
                    ), "Momentum should have been created on round 1."
                    self.momentum_vector = [
                        self.server_momentum * x + y
                        for x, y in zip(
                            self.momentum_vector, pseudo_gradient, strict=True
                        )
                    ]
                else:
                    self.momentum_vector = pseudo_gradient

                # No nesterov for now
                pseudo_gradient = self.momentum_vector

            # SGD
            fedavg_result = [
                x - self.server_learning_rate * y
                for x, y in zip(initial_weights, pseudo_gradient, strict=True)
            ]
            # Update current weights
            self.initial_parameters = ndarrays_to_parameters(fedavg_result)

        parameters_aggregated = ndarrays_to_parameters(fedavg_result)

        # Aggregate custom metrics if aggregation fn was provided
        metrics_aggregated = {}
        if self.fit_metrics_aggregation_fn:
            fit_metrics = [(res.num_examples, res.metrics) for _, res in results]
            metrics_aggregated = self.fit_metrics_aggregation_fn(fit_metrics)
        elif server_round == 1:  # Only log this warning once
            log(WARNING, "No fit_metrics_aggregation_fn provided")

        return parameters_aggregated, metrics_aggregated



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower SQLite ObjectStore implementation."""


from typing import cast

from flwr.common.inflatable import (
    get_object_id,
    is_valid_sha256_hash,
    iterate_object_tree,
)
from flwr.common.inflatable_utils import validate_object_content
from flwr.proto.message_pb2 import ObjectTree  # pylint: disable=E0611
from flwr.supercore.sqlite_mixin import SqliteMixin
from flwr.supercore.utils import uint64_to_int64

from .object_store import NoObjectInStoreError, ObjectStore

SQL_CREATE_OBJECTS = """
CREATE TABLE IF NOT EXISTS objects (
    object_id               TEXT PRIMARY KEY,
    content                 BLOB,
    is_available            INTEGER NOT NULL CHECK (is_available IN (0,1)),
    ref_count               INTEGER NOT NULL
);
"""
SQL_CREATE_OBJECT_CHILDREN = """
CREATE TABLE IF NOT EXISTS object_children (
    parent_id               TEXT NOT NULL,
    child_id                TEXT NOT NULL,
    FOREIGN KEY (parent_id) REFERENCES objects(object_id) ON DELETE CASCADE,
    FOREIGN KEY (child_id)  REFERENCES objects(object_id) ON DELETE CASCADE,
    PRIMARY KEY (parent_id, child_id)
);
"""
SQL_CREATE_RUN_OBJECTS = """
CREATE TABLE IF NOT EXISTS run_objects (
    run_id                  INTEGER NOT NULL,
    object_id               TEXT NOT NULL,
    FOREIGN KEY (object_id) REFERENCES objects(object_id) ON DELETE CASCADE,
    PRIMARY KEY (run_id, object_id)
);
"""


class SqliteObjectStore(ObjectStore, SqliteMixin):
    """SQLite-based implementation of the ObjectStore interface."""

    def __init__(self, database_path: str, verify: bool = True) -> None:
        super().__init__(database_path)
        self.verify = verify

    def get_sql_statements(self) -> tuple[str, ...]:
        """Return SQL statements for ObjectStore tables."""
        return (
            SQL_CREATE_OBJECTS,
            SQL_CREATE_OBJECT_CHILDREN,
            SQL_CREATE_RUN_OBJECTS,
        )

    def preregister(self, run_id: int, object_tree: ObjectTree) -> list[str]:
        """Identify and preregister missing objects in the `ObjectStore`."""
        new_objects = []
        for tree_node in iterate_object_tree(object_tree):
            obj_id = tree_node.object_id
            if not is_valid_sha256_hash(obj_id):
                raise ValueError(f"Invalid object ID format: {obj_id}")

            child_ids = [child.object_id for child in tree_node.children]
            with self.conn:
                row = self.conn.execute(
                    "SELECT object_id, is_available FROM objects WHERE object_id=?",
                    (obj_id,),
                ).fetchone()
                if row is None:
                    # Insert new object
                    self.conn.execute(
                        "INSERT INTO objects"
                        "(object_id, content, is_available, ref_count) "
                        "VALUES (?, ?, ?, ?)",
                        (obj_id, b"", 0, 0),
                    )
                    for cid in child_ids:
                        self.conn.execute(
                            "INSERT INTO object_children(parent_id, child_id) "
                            "VALUES (?, ?)",
                            (obj_id, cid),
                        )
                        self.conn.execute(
                            "UPDATE objects SET ref_count = ref_count + 1 "
                            "WHERE object_id = ?",
                            (cid,),
                        )
                    new_objects.append(obj_id)
                else:
                    # Add to the list of new objects if not available
                    if not row["is_available"]:
                        new_objects.append(obj_id)

                # Ensure run mapping
                self.conn.execute(
                    "INSERT OR IGNORE INTO run_objects(run_id, object_id) "
                    "VALUES (?, ?)",
                    (uint64_to_int64(run_id), obj_id),
                )
        return new_objects

    def get_object_tree(self, object_id: str) -> ObjectTree:
        """Get the object tree for a given object ID."""
        with self.conn:
            row = self.conn.execute(
                "SELECT object_id FROM objects WHERE object_id=?", (object_id,)
            ).fetchone()
            if not row:
                raise NoObjectInStoreError(
                    f"Object {object_id} was not pre-registered."
                )
            children = self.query(
                "SELECT child_id FROM object_children WHERE parent_id=?", (object_id,)
            )

            # Build the object trees of all children
            try:
                child_trees = [self.get_object_tree(ch["child_id"]) for ch in children]
            except NoObjectInStoreError as e:
                # Raise an error if any child object is missing
                # This indicates an integrity issue
                raise NoObjectInStoreError(
                    f"Object tree for object ID '{object_id}' contains missing "
                    "children. This may indicate a corrupted object store."
                ) from e

            # Create and return the ObjectTree for the current object
            return ObjectTree(object_id=object_id, children=child_trees)

    def put(self, object_id: str, object_content: bytes) -> None:
        """Put an object into the store."""
        if self.verify:
            # Verify object_id and object_content match
            object_id_from_content = get_object_id(object_content)
            if object_id != object_id_from_content:
                raise ValueError(f"Object ID {object_id} does not match content hash")

            # Validate object content
            validate_object_content(content=object_content)

        with self.conn:
            # Only allow adding the object if it has been preregistered
            row = self.conn.execute(
                "SELECT is_available FROM objects WHERE object_id=?", (object_id,)
            ).fetchone()
            if row is None:
                raise NoObjectInStoreError(
                    f"Object with ID '{object_id}' was not pre-registered."
                )

            # Return if object is already present in the store
            if row["is_available"]:
                return

            # Update the object entry in the store
            self.conn.execute(
                "UPDATE objects SET content=?, is_available=1 WHERE object_id=?",
                (object_content, object_id),
            )

    def get(self, object_id: str) -> bytes | None:
        """Get an object from the store."""
        rows = self.query("SELECT content FROM objects WHERE object_id=?", (object_id,))
        return rows[0]["content"] if rows else None

    def delete(self, object_id: str) -> None:
        """Delete an object and its unreferenced descendants from the store."""
        with self.conn:
            row = self.conn.execute(
                "SELECT ref_count FROM objects WHERE object_id=?", (object_id,)
            ).fetchone()

            # If the object is not in the store, nothing to delete
            if row is None:
                return

            # Skip deletion if there are still references
            if row["ref_count"] > 0:
                return

            # Deleting will cascade via FK, but we need to decrement children first
            children = self.conn.execute(
                "SELECT child_id FROM object_children WHERE parent_id=?", (object_id,)
            ).fetchall()
            child_ids = [child["child_id"] for child in children]

            if child_ids:
                placeholders = ", ".join("?" for _ in child_ids)
                query = f"""
                    UPDATE objects SET ref_count = ref_count - 1
                    WHERE object_id IN ({placeholders})
                """
                self.conn.execute(query, child_ids)

            self.conn.execute("DELETE FROM objects WHERE object_id=?", (object_id,))

            # Recursively clean children
            for child_id in child_ids:
                self.delete(child_id)

    def delete_objects_in_run(self, run_id: int) -> None:
        """Delete all objects that were registered in a specific run."""
        run_id_sint = uint64_to_int64(run_id)
        with self.conn:
            objs = self.conn.execute(
                "SELECT object_id FROM run_objects WHERE run_id=?", (run_id_sint,)
            ).fetchall()
            for obj in objs:
                object_id = obj["object_id"]
                row = self.conn.execute(
                    "SELECT ref_count FROM objects WHERE object_id=?", (object_id,)
                ).fetchone()
                if row and row["ref_count"] == 0:
                    self.delete(object_id)
            self.conn.execute("DELETE FROM run_objects WHERE run_id=?", (run_id_sint,))

    def clear(self) -> None:
        """Clear the store."""
        with self.conn:
            self.conn.execute("DELETE FROM object_children;")
            self.conn.execute("DELETE FROM run_objects;")
            self.conn.execute("DELETE FROM objects;")

    def __contains__(self, object_id: str) -> bool:
        """Check if an object_id is in the store."""
        row = self.conn.execute(
            "SELECT 1 FROM objects WHERE object_id=?", (object_id,)
        ).fetchone()
        return row is not None

    def __len__(self) -> int:
        """Return the number of objects in the store."""
        row = self.conn.execute("SELECT COUNT(*) AS cnt FROM objects;").fetchone()
        return cast(int, row["cnt"])



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface `supernode unregister` command."""


import io
import json
from pathlib import Path
from typing import Annotated

import typer
from rich.console import Console

from flwr.cli.config_utils import (
    exit_if_no_address,
    load_and_validate,
    process_loaded_project_config,
    validate_federation_in_project_config,
)
from flwr.common.constant import FAB_CONFIG_FILE, CliOutputFormat
from flwr.common.logger import print_json_error, redirect_output, restore_output
from flwr.proto.control_pb2 import UnregisterNodeRequest  # pylint: disable=E0611
from flwr.proto.control_pb2_grpc import ControlStub

from ..utils import flwr_cli_grpc_exc_handler, init_channel, load_cli_auth_plugin


def unregister(  # pylint: disable=R0914
    node_id: Annotated[
        int,
        typer.Argument(
            help="ID of the SuperNode to remove.",
        ),
    ],
    app: Annotated[
        Path,
        typer.Argument(help="Path of the Flower project"),
    ] = Path("."),
    federation: Annotated[
        str | None,
        typer.Argument(help="Name of the federation"),
    ] = None,
    output_format: Annotated[
        str,
        typer.Option(
            "--format",
            case_sensitive=False,
            help="Format output using 'default' view or 'json'",
        ),
    ] = CliOutputFormat.DEFAULT,
) -> None:
    """Unregister a SuperNode from the federation."""
    suppress_output = output_format == CliOutputFormat.JSON
    captured_output = io.StringIO()

    try:
        if suppress_output:
            redirect_output(captured_output)

        # Load and validate federation config
        typer.secho("Loading project configuration... ", fg=typer.colors.BLUE)

        pyproject_path = app / FAB_CONFIG_FILE if app else None
        config, errors, warnings = load_and_validate(pyproject_path, check_module=False)
        config = process_loaded_project_config(config, errors, warnings)
        federation, federation_config = validate_federation_in_project_config(
            federation, config
        )
        exit_if_no_address(federation_config, "supernode unregister")

        channel = None
        try:
            auth_plugin = load_cli_auth_plugin(app, federation, federation_config)
            channel = init_channel(app, federation_config, auth_plugin)
            stub = ControlStub(channel)  # pylint: disable=unused-variable # noqa: F841

            _unregister_node(stub=stub, node_id=node_id, output_format=output_format)

        except ValueError as err:
            typer.secho(
                f"❌ {err}",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            raise typer.Exit(code=1) from err
        finally:
            if channel:
                channel.close()

    except (typer.Exit, Exception) as err:  # pylint: disable=broad-except
        if suppress_output:
            restore_output()
            e_message = captured_output.getvalue()
            print_json_error(e_message, err)
        else:
            typer.secho(
                f"{err}",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
    finally:
        if suppress_output:
            restore_output()
        captured_output.close()


def _unregister_node(
    stub: ControlStub,
    node_id: int,
    output_format: str,
) -> None:
    """Unregister a SuperNode from the federation."""
    with flwr_cli_grpc_exc_handler():
        stub.UnregisterNode(request=UnregisterNodeRequest(node_id=node_id))
    typer.secho(
        f"✅ SuperNode {node_id} unregistered successfully.", fg=typer.colors.GREEN
    )
    if output_format == CliOutputFormat.JSON:
        run_output = json.dumps(
            {
                "success": True,
                "node-id": node_id,
            }
        )
        restore_output()
        Console().print_json(run_output)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import abc
import collections.abc
import flwr.proto.control_pb2
import grpc
import grpc.aio
import typing

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta): ...

class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

class ControlStub:
    def __init__(self, channel: typing.Union[grpc.Channel, grpc.aio.Channel]) -> None: ...
    StartRun: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.StartRunRequest,
        flwr.proto.control_pb2.StartRunResponse,
    ]
    """Start run upon request"""

    StopRun: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.StopRunRequest,
        flwr.proto.control_pb2.StopRunResponse,
    ]
    """Stop run upon request"""

    StreamLogs: grpc.UnaryStreamMultiCallable[
        flwr.proto.control_pb2.StreamLogsRequest,
        flwr.proto.control_pb2.StreamLogsResponse,
    ]
    """Start log stream upon request"""

    ListRuns: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ListRunsRequest,
        flwr.proto.control_pb2.ListRunsResponse,
    ]
    """flwr ls command"""

    GetLoginDetails: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.GetLoginDetailsRequest,
        flwr.proto.control_pb2.GetLoginDetailsResponse,
    ]
    """Get login details upon request"""

    GetAuthTokens: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.GetAuthTokensRequest,
        flwr.proto.control_pb2.GetAuthTokensResponse,
    ]
    """Get auth tokens upon request"""

    PullArtifacts: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.PullArtifactsRequest,
        flwr.proto.control_pb2.PullArtifactsResponse,
    ]
    """Pull artifacts generated during a run (flwr pull)"""

    RegisterNode: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.RegisterNodeRequest,
        flwr.proto.control_pb2.RegisterNodeResponse,
    ]
    """Register SuperNode"""

    UnregisterNode: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.UnregisterNodeRequest,
        flwr.proto.control_pb2.UnregisterNodeResponse,
    ]
    """Unregister SuperNode"""

    ListNodes: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ListNodesRequest,
        flwr.proto.control_pb2.ListNodesResponse,
    ]
    """List SuperNodes"""

    ListFederations: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ListFederationsRequest,
        flwr.proto.control_pb2.ListFederationsResponse,
    ]
    """List Federations"""

    ShowFederation: grpc.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ShowFederationRequest,
        flwr.proto.control_pb2.ShowFederationResponse,
    ]
    """Show Federation"""

class ControlAsyncStub:
    StartRun: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.StartRunRequest,
        flwr.proto.control_pb2.StartRunResponse,
    ]
    """Start run upon request"""

    StopRun: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.StopRunRequest,
        flwr.proto.control_pb2.StopRunResponse,
    ]
    """Stop run upon request"""

    StreamLogs: grpc.aio.UnaryStreamMultiCallable[
        flwr.proto.control_pb2.StreamLogsRequest,
        flwr.proto.control_pb2.StreamLogsResponse,
    ]
    """Start log stream upon request"""

    ListRuns: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ListRunsRequest,
        flwr.proto.control_pb2.ListRunsResponse,
    ]
    """flwr ls command"""

    GetLoginDetails: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.GetLoginDetailsRequest,
        flwr.proto.control_pb2.GetLoginDetailsResponse,
    ]
    """Get login details upon request"""

    GetAuthTokens: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.GetAuthTokensRequest,
        flwr.proto.control_pb2.GetAuthTokensResponse,
    ]
    """Get auth tokens upon request"""

    PullArtifacts: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.PullArtifactsRequest,
        flwr.proto.control_pb2.PullArtifactsResponse,
    ]
    """Pull artifacts generated during a run (flwr pull)"""

    RegisterNode: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.RegisterNodeRequest,
        flwr.proto.control_pb2.RegisterNodeResponse,
    ]
    """Register SuperNode"""

    UnregisterNode: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.UnregisterNodeRequest,
        flwr.proto.control_pb2.UnregisterNodeResponse,
    ]
    """Unregister SuperNode"""

    ListNodes: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ListNodesRequest,
        flwr.proto.control_pb2.ListNodesResponse,
    ]
    """List SuperNodes"""

    ListFederations: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ListFederationsRequest,
        flwr.proto.control_pb2.ListFederationsResponse,
    ]
    """List Federations"""

    ShowFederation: grpc.aio.UnaryUnaryMultiCallable[
        flwr.proto.control_pb2.ShowFederationRequest,
        flwr.proto.control_pb2.ShowFederationResponse,
    ]
    """Show Federation"""

class ControlServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def StartRun(
        self,
        request: flwr.proto.control_pb2.StartRunRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.StartRunResponse, collections.abc.Awaitable[flwr.proto.control_pb2.StartRunResponse]]:
        """Start run upon request"""

    @abc.abstractmethod
    def StopRun(
        self,
        request: flwr.proto.control_pb2.StopRunRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.StopRunResponse, collections.abc.Awaitable[flwr.proto.control_pb2.StopRunResponse]]:
        """Stop run upon request"""

    @abc.abstractmethod
    def StreamLogs(
        self,
        request: flwr.proto.control_pb2.StreamLogsRequest,
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[flwr.proto.control_pb2.StreamLogsResponse], collections.abc.AsyncIterator[flwr.proto.control_pb2.StreamLogsResponse]]:
        """Start log stream upon request"""

    @abc.abstractmethod
    def ListRuns(
        self,
        request: flwr.proto.control_pb2.ListRunsRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.ListRunsResponse, collections.abc.Awaitable[flwr.proto.control_pb2.ListRunsResponse]]:
        """flwr ls command"""

    @abc.abstractmethod
    def GetLoginDetails(
        self,
        request: flwr.proto.control_pb2.GetLoginDetailsRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.GetLoginDetailsResponse, collections.abc.Awaitable[flwr.proto.control_pb2.GetLoginDetailsResponse]]:
        """Get login details upon request"""

    @abc.abstractmethod
    def GetAuthTokens(
        self,
        request: flwr.proto.control_pb2.GetAuthTokensRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.GetAuthTokensResponse, collections.abc.Awaitable[flwr.proto.control_pb2.GetAuthTokensResponse]]:
        """Get auth tokens upon request"""

    @abc.abstractmethod
    def PullArtifacts(
        self,
        request: flwr.proto.control_pb2.PullArtifactsRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.PullArtifactsResponse, collections.abc.Awaitable[flwr.proto.control_pb2.PullArtifactsResponse]]:
        """Pull artifacts generated during a run (flwr pull)"""

    @abc.abstractmethod
    def RegisterNode(
        self,
        request: flwr.proto.control_pb2.RegisterNodeRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.RegisterNodeResponse, collections.abc.Awaitable[flwr.proto.control_pb2.RegisterNodeResponse]]:
        """Register SuperNode"""

    @abc.abstractmethod
    def UnregisterNode(
        self,
        request: flwr.proto.control_pb2.UnregisterNodeRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.UnregisterNodeResponse, collections.abc.Awaitable[flwr.proto.control_pb2.UnregisterNodeResponse]]:
        """Unregister SuperNode"""

    @abc.abstractmethod
    def ListNodes(
        self,
        request: flwr.proto.control_pb2.ListNodesRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.ListNodesResponse, collections.abc.Awaitable[flwr.proto.control_pb2.ListNodesResponse]]:
        """List SuperNodes"""

    @abc.abstractmethod
    def ListFederations(
        self,
        request: flwr.proto.control_pb2.ListFederationsRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.ListFederationsResponse, collections.abc.Awaitable[flwr.proto.control_pb2.ListFederationsResponse]]:
        """List Federations"""

    @abc.abstractmethod
    def ShowFederation(
        self,
        request: flwr.proto.control_pb2.ShowFederationRequest,
        context: _ServicerContext,
    ) -> typing.Union[flwr.proto.control_pb2.ShowFederationResponse, collections.abc.Awaitable[flwr.proto.control_pb2.ShowFederationResponse]]:
        """Show Federation"""

def add_ControlServicer_to_server(servicer: ControlServicer, server: typing.Union[grpc.Server, grpc.aio.Server]) -> None: ...



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface utils."""


import hashlib
import json
import re
from collections.abc import Callable, Iterable, Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, cast

import grpc
import pathspec
import requests
import typer

from flwr.common.constant import (
    ACCESS_TOKEN_KEY,
    AUTHN_TYPE_JSON_KEY,
    CREDENTIALS_DIR,
    FLWR_DIR,
    NO_ACCOUNT_AUTH_MESSAGE,
    NO_ARTIFACT_PROVIDER_MESSAGE,
    NODE_NOT_FOUND_MESSAGE,
    PUBLIC_KEY_ALREADY_IN_USE_MESSAGE,
    PUBLIC_KEY_NOT_VALID,
    PULL_UNFINISHED_RUN_MESSAGE,
    REFRESH_TOKEN_KEY,
    RUN_ID_NOT_FOUND_MESSAGE,
    AuthnType,
)
from flwr.common.grpc import (
    GRPC_MAX_MESSAGE_LENGTH,
    create_channel,
    on_channel_state_change,
)
from flwr.common.version import package_version as flwr_version
from flwr.supercore.constant import APP_ID_PATTERN, APP_VERSION_PATTERN

from .auth_plugin import CliAuthPlugin, get_cli_plugin_class
from .cli_account_auth_interceptor import CliAccountAuthInterceptor
from .config_utils import validate_certificate_in_federation_config


def prompt_text(
    text: str,
    predicate: Callable[[str], bool] = lambda _: True,
    default: str | None = None,
) -> str:
    """Ask user to enter text input.

    Parameters
    ----------
    text : str
        The prompt text to display to the user.
    predicate : Callable[[str], bool] (default: lambda _: True)
        A function to validate the user input. Default accepts all non-empty strings.
    default : str | None (default: None)
        Default value to use if user presses enter without input.

    Returns
    -------
    str
        The validated user input.
    """
    while True:
        result = typer.prompt(
            typer.style(f"\n💬 {text}", fg=typer.colors.MAGENTA, bold=True),
            default=default,
        )
        if predicate(result) and len(result) > 0:
            break
        print(typer.style("❌ Invalid entry", fg=typer.colors.RED, bold=True))

    return cast(str, result)


def prompt_options(text: str, options: list[str]) -> str:
    """Ask user to select one of the given options and return the selected item.

    Parameters
    ----------
    text : str
        The prompt text to display to the user.
    options : list[str]
        List of options to present to the user.

    Returns
    -------
    str
        The selected option from the list.
    """
    # Turn options into a list with index as in " [ 0] quickstart-pytorch"
    options_formatted = [
        " [ "
        + typer.style(index, fg=typer.colors.GREEN, bold=True)
        + "]"
        + f" {typer.style(name, fg=typer.colors.WHITE, bold=True)}"
        for index, name in enumerate(options)
    ]

    while True:
        index = typer.prompt(
            "\n"
            + typer.style(f"💬 {text}", fg=typer.colors.MAGENTA, bold=True)
            + "\n\n"
            + "\n".join(options_formatted)
            + "\n\n\n"
        )
        try:
            options[int(index)]  # pylint: disable=expression-not-assigned
            break
        except IndexError:
            print(typer.style("❌ Index out of range", fg=typer.colors.RED, bold=True))
            continue
        except ValueError:
            print(
                typer.style("❌ Please choose a number", fg=typer.colors.RED, bold=True)
            )
            continue

    result = options[int(index)]
    return result


def is_valid_project_name(name: str) -> bool:
    """Check if the given string is a valid Python project name.

    A valid project name must start with a letter and can only contain letters, digits,
    and hyphens.
    """
    if not name:
        return False

    # Check if the first character is a letter
    if not name[0].isalpha():
        return False

    # Check if the rest of the characters are valid (letter, digit, or dash)
    for char in name[1:]:
        if not (char.isalnum() or char in "-"):
            return False

    return True


def sanitize_project_name(name: str) -> str:
    """Sanitize the given string to make it a valid Python project name.

    This function replaces spaces, dots, slashes, and underscores with dashes, removes
    any characters not allowed in Python project names, makes the string lowercase, and
    ensures it starts with a valid character.

    Parameters
    ----------
    name : str
        The project name to sanitize.

    Returns
    -------
    str
        The sanitized project name that is valid for Python projects.
    """
    # Replace whitespace with '_'
    name_with_hyphens = re.sub(r"[ ./_]", "-", name)

    # Allowed characters in a module name: letters, digits, underscore
    allowed_chars = set(
        "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-"
    )

    # Make the string lowercase
    sanitized_name = name_with_hyphens.lower()

    # Remove any characters not allowed in Python module names
    sanitized_name = "".join(c for c in sanitized_name if c in allowed_chars)

    # Ensure the first character is a letter or underscore
    while sanitized_name and (
        sanitized_name[0].isdigit() or sanitized_name[0] not in allowed_chars
    ):
        sanitized_name = sanitized_name[1:]

    return sanitized_name


def get_sha256_hash(file_path_or_int: Path | int) -> str:
    """Calculate the SHA-256 hash of a file or integer.

    Parameters
    ----------
    file_path_or_int : Path | int
        Either a path to a file to hash, or an integer to convert to string and hash.

    Returns
    -------
    str
        The SHA-256 hash as a hexadecimal string.
    """
    sha256 = hashlib.sha256()
    if isinstance(file_path_or_int, Path):
        with open(file_path_or_int, "rb") as f:
            while True:
                data = f.read(65536)  # Read in 64kB blocks
                if not data:
                    break
                sha256.update(data)
    elif isinstance(file_path_or_int, int):
        sha256.update(str(file_path_or_int).encode())
    return sha256.hexdigest()


def get_account_auth_config_path(root_dir: Path, federation: str) -> Path:
    """Return the path to the account auth config file.

    Additionally, a `.gitignore` file will be created in the Flower directory to
    include the `.credentials` folder to be excluded from git. If the `.gitignore`
    file already exists, a warning will be displayed if the `.credentials` entry is
    not found.
    """
    # Locate the credentials directory
    abs_flwr_dir = root_dir.absolute() / FLWR_DIR
    credentials_dir = abs_flwr_dir / CREDENTIALS_DIR
    credentials_dir.mkdir(parents=True, exist_ok=True)

    # Determine the absolute path of the Flower directory for .gitignore
    gitignore_path = abs_flwr_dir / ".gitignore"
    credential_entry = CREDENTIALS_DIR

    try:
        if gitignore_path.exists():
            with open(gitignore_path, encoding="utf-8") as gitignore_file:
                lines = gitignore_file.read().splitlines()

            # Warn if .credentials is not already in .gitignore
            if credential_entry not in lines:
                typer.secho(
                    f"`.gitignore` exists, but `{credential_entry}` entry not found. "
                    "Consider adding it to your `.gitignore` to exclude Flower "
                    "credentials from git.",
                    fg=typer.colors.YELLOW,
                    bold=True,
                )
        else:
            typer.secho(
                f"Creating a new `.gitignore` with `{credential_entry}` entry...",
                fg=typer.colors.BLUE,
            )
            # Create a new .gitignore with .credentials
            with open(gitignore_path, "w", encoding="utf-8") as gitignore_file:
                gitignore_file.write(f"{credential_entry}\n")
    except Exception as err:
        typer.secho(
            "❌ An error occurred while handling `.gitignore.` "
            f"Please check the permissions of `{gitignore_path}` and try again.",
            fg=typer.colors.RED,
            bold=True,
            err=True,
        )
        raise typer.Exit(code=1) from err

    return credentials_dir / f"{federation}.json"


def account_auth_enabled(federation_config: dict[str, Any]) -> bool:
    """Check if account authentication is enabled in the federation config.

    Parameters
    ----------
    federation_config : dict[str, Any]
        The federation configuration dictionary.

    Returns
    -------
    bool
        True if account authentication is enabled, False otherwise.
    """
    enabled: bool = federation_config.get("enable-user-auth", False)
    enabled |= federation_config.get("enable-account-auth", False)
    if "enable-user-auth" in federation_config:
        typer.secho(
            "`enable-user-auth` is deprecated and will be removed in a future "
            "release. Please use `enable-account-auth` instead.",
            fg=typer.colors.YELLOW,
            bold=True,
        )
    return enabled


def retrieve_authn_type(config_path: Path) -> str:
    """Retrieve the auth type from the config file or return NOOP if not found.

    Parameters
    ----------
    config_path : Path
        Path to the authentication configuration file.

    Returns
    -------
    str
        The authentication type string, or AuthnType.NOOP if not found.
    """
    try:
        with config_path.open("r", encoding="utf-8") as file:
            json_file = json.load(file)
        authn_type: str = json_file[AUTHN_TYPE_JSON_KEY]
        return authn_type
    except (FileNotFoundError, KeyError):
        return AuthnType.NOOP


def load_cli_auth_plugin(
    root_dir: Path,
    federation: str,
    federation_config: dict[str, Any],
    authn_type: str | None = None,
) -> CliAuthPlugin:
    """Load the CLI-side account auth plugin for the given authn type.

    Parameters
    ----------
    root_dir : Path
        Root directory of the Flower project.
    federation : str
        Name of the federation.
    federation_config : dict[str, Any]
        Federation configuration dictionary.
    authn_type : str | None
        Authentication type. If None, will be determined from config.

    Returns
    -------
    CliAuthPlugin
        The loaded authentication plugin instance.

    Raises
    ------
    typer.Exit
        If the authentication type is unknown.
    """
    # Find the path to the account auth config file
    config_path = get_account_auth_config_path(root_dir, federation)

    # Determine the auth type if not provided
    # Only `flwr login` command can provide `authn_type` explicitly, as it can query the
    # SuperLink for the auth type.
    if authn_type is None:
        authn_type = AuthnType.NOOP
        if account_auth_enabled(federation_config):
            authn_type = retrieve_authn_type(config_path)

    # Retrieve auth plugin class and instantiate it
    try:
        auth_plugin_class = get_cli_plugin_class(authn_type)
        return auth_plugin_class(config_path)
    except ValueError:
        typer.echo(f"❌ Unknown account authentication type: {authn_type}")
        raise typer.Exit(code=1) from None


def init_channel(
    app: Path, federation_config: dict[str, Any], auth_plugin: CliAuthPlugin
) -> grpc.Channel:
    """Initialize gRPC channel to the Control API.

    Parameters
    ----------
    app : Path
        Path to the Flower app directory.
    federation_config : dict[str, Any]
        Federation configuration dictionary containing address and TLS settings.
    auth_plugin : CliAuthPlugin
        Authentication plugin instance for handling credentials.

    Returns
    -------
    grpc.Channel
        Configured gRPC channel with authentication interceptors.
    """
    insecure, root_certificates_bytes = validate_certificate_in_federation_config(
        app, federation_config
    )

    # Load tokens
    auth_plugin.load_tokens()

    # Create the gRPC channel
    channel = create_channel(
        server_address=federation_config["address"],
        insecure=insecure,
        root_certificates=root_certificates_bytes,
        max_message_length=GRPC_MAX_MESSAGE_LENGTH,
        interceptors=[CliAccountAuthInterceptor(auth_plugin)],
    )
    channel.subscribe(on_channel_state_change)
    return channel


@contextmanager
def flwr_cli_grpc_exc_handler() -> Iterator[None]:  # pylint: disable=too-many-branches
    """Context manager to handle specific gRPC errors.

    Catches grpc.RpcError exceptions with UNAUTHENTICATED, UNIMPLEMENTED,
    UNAVAILABLE, PERMISSION_DENIED, NOT_FOUND, and FAILED_PRECONDITION statuses,
    informs the user, and exits the application. All other exceptions will be
    allowed to escape.

    Yields
    ------
    None
        Context manager yields nothing.

    Raises
    ------
    typer.Exit
        On handled gRPC error statuses with appropriate exit code.
    grpc.RpcError
        For unhandled gRPC error statuses.
    """
    try:
        yield
    except grpc.RpcError as e:
        if e.code() == grpc.StatusCode.UNAUTHENTICATED:
            typer.secho(
                "❌ Authentication failed. Please run `flwr login`"
                " to authenticate and try again.",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            raise typer.Exit(code=1) from None
        if e.code() == grpc.StatusCode.UNIMPLEMENTED:
            if e.details() == NO_ACCOUNT_AUTH_MESSAGE:  # pylint: disable=E1101
                typer.secho(
                    "❌ Account authentication is not enabled on this SuperLink.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
            elif e.details() == NO_ARTIFACT_PROVIDER_MESSAGE:  # pylint: disable=E1101
                typer.secho(
                    "❌ The SuperLink does not support `flwr pull` command.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
            else:
                typer.secho(
                    "❌ The SuperLink cannot process this request. Please verify that "
                    "you set the address to its Control API endpoint correctly in your "
                    "`pyproject.toml`, and ensure that the Flower versions used by "
                    "the CLI and SuperLink are compatible.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
            raise typer.Exit(code=1) from None
        if e.code() == grpc.StatusCode.PERMISSION_DENIED:
            typer.secho(
                "❌ Permission denied.",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            # pylint: disable-next=E1101
            typer.secho(e.details(), fg=typer.colors.RED, bold=True)
            raise typer.Exit(code=1) from None
        if e.code() == grpc.StatusCode.UNAVAILABLE:
            typer.secho(
                "Connection to the SuperLink is unavailable. Please check your network "
                "connection and 'address' in the federation configuration.",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            raise typer.Exit(code=1) from None
        if e.code() == grpc.StatusCode.NOT_FOUND:
            if e.details() == RUN_ID_NOT_FOUND_MESSAGE:  # pylint: disable=E1101
                typer.secho(
                    "❌ Run ID not found.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
                raise typer.Exit(code=1) from None
            if e.details() == NODE_NOT_FOUND_MESSAGE:  # pylint: disable=E1101
                typer.secho(
                    "❌ Node ID not found for this account.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
                raise typer.Exit(code=1) from None
        if e.code() == grpc.StatusCode.FAILED_PRECONDITION:
            if e.details() == PULL_UNFINISHED_RUN_MESSAGE:  # pylint: disable=E1101
                typer.secho(
                    "❌ Run is not finished yet. Artifacts can only be pulled after "
                    "the run is finished. You can check the run status with `flwr ls`.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
                raise typer.Exit(code=1) from None
            if (
                e.details() == PUBLIC_KEY_ALREADY_IN_USE_MESSAGE
            ):  # pylint: disable=E1101
                typer.secho(
                    "❌ The provided public key is already in use by another "
                    "SuperNode.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
                raise typer.Exit(code=1) from None
            if e.details() == PUBLIC_KEY_NOT_VALID:  # pylint: disable=E1101
                typer.secho(
                    "❌ The provided public key is invalid. Please provide a valid "
                    "NIST EC public key.",
                    fg=typer.colors.RED,
                    bold=True,
                    err=True,
                )
                raise typer.Exit(code=1) from None

            # Log details from grpc error directly
            typer.secho(
                f"❌ {e.details()}",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            raise typer.Exit(code=1) from None
        raise


def request_download_link(
    app_id: str, app_version: str | None, in_url: str, out_url: str
) -> str:
    """Request a download link for the given app from the Flower platform API.

    Parameters
    ----------
    app_id : str
        The application identifier.
    app_version : str | None
        The application version, or None for latest.
    in_url : str
        The API endpoint URL.
    out_url : str
        The key name for the download URL in the response.

    Returns
    -------
    str
        The download URL for the application.

    Raises
    ------
    typer.Exit
        If connection fails, app not found, or API request fails.
    """
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
    }
    body = {
        "app_id": app_id,  # send raw string of app_id
        "app_version": app_version,
        "flwr_version": flwr_version,
    }
    try:
        resp = requests.post(in_url, headers=headers, data=json.dumps(body), timeout=20)
    except requests.RequestException as e:
        typer.secho(
            f"Unable to connect to Platform API: {e}",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1) from e

    if resp.status_code == 404:
        error_message = resp.json()["detail"]
        if isinstance(error_message, dict):
            available_app_versions = error_message["available_app_versions"]
            available_versions_str = (
                ", ".join(map(str, available_app_versions))
                if available_app_versions
                else "None"
            )
            typer.secho(
                f"{app_id}=={app_version} not found in Platform API. "
                f"Available app versions for {app_id}: {available_versions_str}",
                fg=typer.colors.RED,
                err=True,
            )
        else:
            typer.secho(
                f"{app_id} not found in Platform API.",
                fg=typer.colors.RED,
                err=True,
            )
        raise typer.Exit(code=1)

    if not resp.ok:
        typer.secho(
            f"Platform API request failed with "
            f"status {resp.status_code}. Details: {resp.text}",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1)

    data = resp.json()
    if out_url not in data:
        typer.secho(
            "Invalid response from Platform API",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1)
    return str(data[out_url])


def build_pathspec(patterns: Iterable[str]) -> pathspec.PathSpec:
    """Build a PathSpec from a list of GitIgnore-style patterns.

    Parameters
    ----------
    patterns : Iterable[str]
        Iterable of GitIgnore-style pattern strings.

    Returns
    -------
    pathspec.PathSpec
        Compiled PathSpec object for pattern matching.
    """
    return pathspec.PathSpec.from_lines("gitwildmatch", patterns)


def load_gitignore_patterns(file: Path | bytes) -> list[str]:
    """Load gitignore patterns from .gitignore file bytes.

    Parameters
    ----------
    file : Path | bytes
        The path to a .gitignore file or its bytes content.

    Returns
    -------
    list[str]
        List of gitignore patterns.
        Returns empty list if content can't be decoded or the file does not exist.
    """
    try:
        if isinstance(file, Path):
            content = file.read_text(encoding="utf-8")
        else:
            content = file.decode("utf-8")
        patterns = [
            line.strip()
            for line in content.splitlines()
            if line.strip() and not line.strip().startswith("#")
        ]
        return patterns
    except (UnicodeDecodeError, OSError):
        return []


def validate_credentials_content(creds_path: Path) -> str:
    """Load and validate the credentials file content.

    Ensures required keys exist:
      - AUTHN_TYPE_JSON_KEY
      - ACCESS_TOKEN_KEY
      - REFRESH_TOKEN_KEY
    """
    try:
        creds: dict[str, str] = json.loads(creds_path.read_text(encoding="utf-8"))
    except (OSError, json.JSONDecodeError) as err:
        typer.secho(
            f"Invalid credentials file at '{creds_path}': {err}",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1) from err

    required_keys = [AUTHN_TYPE_JSON_KEY, ACCESS_TOKEN_KEY, REFRESH_TOKEN_KEY]
    missing = [key for key in required_keys if key not in creds]

    if missing:
        typer.secho(
            f"Credentials file '{creds_path}' is missing "
            f"required key(s): {', '.join(missing)}. Please log in again.",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1)

    return creds[ACCESS_TOKEN_KEY]


def parse_app_spec(app_spec: str) -> tuple[str, str | None]:
    """Parse app specification string into app ID and version.

    Parameters
    ----------
    app_spec : str
        The app specification string in the format '@account/app' or
        '@account/app==x.y.z' (digits only).

    Returns
    -------
    tuple[str, str | None]
        A tuple containing the app ID and optional version.

    Raises
    ------
    typer.Exit
        If the app specification format is invalid.
    """
    if "==" in app_spec:
        app_id, app_version = app_spec.split("==")

        # Validate app version format
        if not re.match(APP_VERSION_PATTERN, app_version):
            typer.secho(
                "❌ Invalid app version. Expected format: x.y.z (digits only).",
                fg=typer.colors.RED,
                err=True,
            )
            raise typer.Exit(code=1)
    else:
        app_id = app_spec
        app_version = None

    # Validate app_id format
    if not re.match(APP_ID_PATTERN, app_id):
        typer.secho(
            "❌ Invalid remote app ID. Expected format: '@account/app'.",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1)

    return app_id, app_version



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower client interceptor."""


from collections.abc import Callable
from typing import Any

import grpc
from cryptography.hazmat.primitives.asymmetric import ec
from google.protobuf.message import Message as GrpcMessage

from flwr.common import now
from flwr.common.constant import PUBLIC_KEY_HEADER, SIGNATURE_HEADER, TIMESTAMP_HEADER
from flwr.supercore.primitives.asymmetric import public_key_to_bytes, sign_message


class NodeAuthClientInterceptor(grpc.UnaryUnaryClientInterceptor):  # type: ignore
    """Client interceptor for node authentication."""

    def __init__(
        self,
        private_key: ec.EllipticCurvePrivateKey,
        public_key: ec.EllipticCurvePublicKey,
    ):
        self.private_key = private_key
        self.public_key_bytes = public_key_to_bytes(public_key)

    def intercept_unary_unary(
        self,
        continuation: Callable[[Any, Any], Any],
        client_call_details: grpc.ClientCallDetails,
        request: GrpcMessage,
    ) -> grpc.Call:
        """Flower client interceptor.

        Intercept unary call from client and add necessary authentication header in the
        RPC metadata.
        """
        metadata = list(client_call_details.metadata or [])

        # Add the public key
        metadata.append((PUBLIC_KEY_HEADER, self.public_key_bytes))

        # Add timestamp
        timestamp = now().isoformat()
        metadata.append((TIMESTAMP_HEADER, timestamp))

        # Sign and add the signature
        signature = sign_message(self.private_key, timestamp.encode("ascii"))
        metadata.append((SIGNATURE_HEADER, signature))

        # Overwrite the metadata
        details = client_call_details._replace(metadata=metadata)

        return continuation(details, request)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import flwr.proto.error_pb2
import flwr.proto.node_pb2
import flwr.proto.recorddict_pb2
import flwr.proto.transport_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Message(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METADATA_FIELD_NUMBER: builtins.int
    CONTENT_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    @property
    def metadata(self) -> global___Metadata: ...
    @property
    def content(self) -> flwr.proto.recorddict_pb2.RecordDict: ...
    @property
    def error(self) -> flwr.proto.error_pb2.Error: ...
    def __init__(
        self,
        *,
        metadata: global___Metadata | None = ...,
        content: flwr.proto.recorddict_pb2.RecordDict | None = ...,
        error: flwr.proto.error_pb2.Error | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["content", b"content", "error", b"error", "metadata", b"metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["content", b"content", "error", b"error", "metadata", b"metadata"]) -> None: ...

global___Message = Message

@typing.final
class Context(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class NodeConfigEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> flwr.proto.transport_pb2.Scalar: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: flwr.proto.transport_pb2.Scalar | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    @typing.final
    class RunConfigEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> flwr.proto.transport_pb2.Scalar: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: flwr.proto.transport_pb2.Scalar | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    RUN_ID_FIELD_NUMBER: builtins.int
    NODE_ID_FIELD_NUMBER: builtins.int
    NODE_CONFIG_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    RUN_CONFIG_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    node_id: builtins.int
    @property
    def node_config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, flwr.proto.transport_pb2.Scalar]: ...
    @property
    def state(self) -> flwr.proto.recorddict_pb2.RecordDict: ...
    @property
    def run_config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, flwr.proto.transport_pb2.Scalar]: ...
    def __init__(
        self,
        *,
        run_id: builtins.int = ...,
        node_id: builtins.int = ...,
        node_config: collections.abc.Mapping[builtins.str, flwr.proto.transport_pb2.Scalar] | None = ...,
        state: flwr.proto.recorddict_pb2.RecordDict | None = ...,
        run_config: collections.abc.Mapping[builtins.str, flwr.proto.transport_pb2.Scalar] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["state", b"state"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["node_config", b"node_config", "node_id", b"node_id", "run_config", b"run_config", "run_id", b"run_id", "state", b"state"]) -> None: ...

global___Context = Context

@typing.final
class Metadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RUN_ID_FIELD_NUMBER: builtins.int
    MESSAGE_ID_FIELD_NUMBER: builtins.int
    SRC_NODE_ID_FIELD_NUMBER: builtins.int
    DST_NODE_ID_FIELD_NUMBER: builtins.int
    REPLY_TO_MESSAGE_ID_FIELD_NUMBER: builtins.int
    GROUP_ID_FIELD_NUMBER: builtins.int
    TTL_FIELD_NUMBER: builtins.int
    MESSAGE_TYPE_FIELD_NUMBER: builtins.int
    CREATED_AT_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    message_id: builtins.str
    src_node_id: builtins.int
    dst_node_id: builtins.int
    reply_to_message_id: builtins.str
    group_id: builtins.str
    ttl: builtins.float
    message_type: builtins.str
    created_at: builtins.float
    def __init__(
        self,
        *,
        run_id: builtins.int = ...,
        message_id: builtins.str = ...,
        src_node_id: builtins.int = ...,
        dst_node_id: builtins.int = ...,
        reply_to_message_id: builtins.str = ...,
        group_id: builtins.str = ...,
        ttl: builtins.float = ...,
        message_type: builtins.str = ...,
        created_at: builtins.float = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["created_at", b"created_at", "dst_node_id", b"dst_node_id", "group_id", b"group_id", "message_id", b"message_id", "message_type", b"message_type", "reply_to_message_id", b"reply_to_message_id", "run_id", b"run_id", "src_node_id", b"src_node_id", "ttl", b"ttl"]) -> None: ...

global___Metadata = Metadata

@typing.final
class ObjectIDs(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OBJECT_IDS_FIELD_NUMBER: builtins.int
    @property
    def object_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        object_ids: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["object_ids", b"object_ids"]) -> None: ...

global___ObjectIDs = ObjectIDs

@typing.final
class ObjectTree(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OBJECT_ID_FIELD_NUMBER: builtins.int
    CHILDREN_FIELD_NUMBER: builtins.int
    object_id: builtins.str
    @property
    def children(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ObjectTree]: ...
    def __init__(
        self,
        *,
        object_id: builtins.str = ...,
        children: collections.abc.Iterable[global___ObjectTree] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["children", b"children", "object_id", b"object_id"]) -> None: ...

global___ObjectTree = ObjectTree

@typing.final
class PushObjectRequest(google.protobuf.message.Message):
    """PushObject messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    OBJECT_ID_FIELD_NUMBER: builtins.int
    OBJECT_CONTENT_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    object_id: builtins.str
    object_content: builtins.bytes
    @property
    def node(self) -> flwr.proto.node_pb2.Node: ...
    def __init__(
        self,
        *,
        node: flwr.proto.node_pb2.Node | None = ...,
        run_id: builtins.int = ...,
        object_id: builtins.str = ...,
        object_content: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["node", b"node"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["node", b"node", "object_content", b"object_content", "object_id", b"object_id", "run_id", b"run_id"]) -> None: ...

global___PushObjectRequest = PushObjectRequest

@typing.final
class PushObjectResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STORED_FIELD_NUMBER: builtins.int
    stored: builtins.bool
    def __init__(
        self,
        *,
        stored: builtins.bool = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["stored", b"stored"]) -> None: ...

global___PushObjectResponse = PushObjectResponse

@typing.final
class PullObjectRequest(google.protobuf.message.Message):
    """PullObject messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    OBJECT_ID_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    object_id: builtins.str
    @property
    def node(self) -> flwr.proto.node_pb2.Node: ...
    def __init__(
        self,
        *,
        node: flwr.proto.node_pb2.Node | None = ...,
        run_id: builtins.int = ...,
        object_id: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["node", b"node"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["node", b"node", "object_id", b"object_id", "run_id", b"run_id"]) -> None: ...

global___PullObjectRequest = PullObjectRequest

@typing.final
class PullObjectResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OBJECT_FOUND_FIELD_NUMBER: builtins.int
    OBJECT_AVAILABLE_FIELD_NUMBER: builtins.int
    OBJECT_CONTENT_FIELD_NUMBER: builtins.int
    object_found: builtins.bool
    object_available: builtins.bool
    object_content: builtins.bytes
    def __init__(
        self,
        *,
        object_found: builtins.bool = ...,
        object_available: builtins.bool = ...,
        object_content: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["object_available", b"object_available", "object_content", b"object_content", "object_found", b"object_found"]) -> None: ...

global___PullObjectResponse = PullObjectResponse

@typing.final
class ConfirmMessageReceivedRequest(google.protobuf.message.Message):
    """ConfirmMessageReceived messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    MESSAGE_OBJECT_ID_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    message_object_id: builtins.str
    @property
    def node(self) -> flwr.proto.node_pb2.Node: ...
    def __init__(
        self,
        *,
        node: flwr.proto.node_pb2.Node | None = ...,
        run_id: builtins.int = ...,
        message_object_id: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["node", b"node"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["message_object_id", b"message_object_id", "node", b"node", "run_id", b"run_id"]) -> None: ...

global___ConfirmMessageReceivedRequest = ConfirmMessageReceivedRequest

@typing.final
class ConfirmMessageReceivedResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ConfirmMessageReceivedResponse = ConfirmMessageReceivedResponse



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: flwr/proto/fleet.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'flwr/proto/fleet.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from flwr.proto import heartbeat_pb2 as flwr_dot_proto_dot_heartbeat__pb2
from flwr.proto import node_pb2 as flwr_dot_proto_dot_node__pb2
from flwr.proto import run_pb2 as flwr_dot_proto_dot_run__pb2
from flwr.proto import fab_pb2 as flwr_dot_proto_dot_fab__pb2
from flwr.proto import message_pb2 as flwr_dot_proto_dot_message__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x16\x66lwr/proto/fleet.proto\x12\nflwr.proto\x1a\x1a\x66lwr/proto/heartbeat.proto\x1a\x15\x66lwr/proto/node.proto\x1a\x14\x66lwr/proto/run.proto\x1a\x14\x66lwr/proto/fab.proto\x1a\x18\x66lwr/proto/message.proto\".\n\x18RegisterNodeFleetRequest\x12\x12\n\npublic_key\x18\x01 \x01(\x0c\",\n\x19RegisterNodeFleetResponse\x12\x0f\n\x07node_id\x18\x01 \x01(\x04\"E\n\x13\x41\x63tivateNodeRequest\x12\x12\n\npublic_key\x18\x01 \x01(\x0c\x12\x1a\n\x12heartbeat_interval\x18\x02 \x01(\x01\"\'\n\x14\x41\x63tivateNodeResponse\x12\x0f\n\x07node_id\x18\x01 \x01(\x04\"(\n\x15\x44\x65\x61\x63tivateNodeRequest\x12\x0f\n\x07node_id\x18\x01 \x01(\x04\"\x18\n\x16\x44\x65\x61\x63tivateNodeResponse\"-\n\x1aUnregisterNodeFleetRequest\x12\x0f\n\x07node_id\x18\x01 \x01(\x04\"\x1d\n\x1bUnregisterNodeFleetResponse\"J\n\x13PullMessagesRequest\x12\x1e\n\x04node\x18\x01 \x01(\x0b\x32\x10.flwr.proto.Node\x12\x13\n\x0bmessage_ids\x18\x02 \x03(\t\"\xa2\x01\n\x14PullMessagesResponse\x12(\n\treconnect\x18\x01 \x01(\x0b\x32\x15.flwr.proto.Reconnect\x12*\n\rmessages_list\x18\x02 \x03(\x0b\x32\x13.flwr.proto.Message\x12\x34\n\x14message_object_trees\x18\x03 \x03(\x0b\x32\x16.flwr.proto.ObjectTree\"\x97\x01\n\x13PushMessagesRequest\x12\x1e\n\x04node\x18\x01 \x01(\x0b\x32\x10.flwr.proto.Node\x12*\n\rmessages_list\x18\x02 \x03(\x0b\x32\x13.flwr.proto.Message\x12\x34\n\x14message_object_trees\x18\x03 \x03(\x0b\x32\x16.flwr.proto.ObjectTree\"\xc9\x01\n\x14PushMessagesResponse\x12(\n\treconnect\x18\x01 \x01(\x0b\x32\x15.flwr.proto.Reconnect\x12>\n\x07results\x18\x02 \x03(\x0b\x32-.flwr.proto.PushMessagesResponse.ResultsEntry\x12\x17\n\x0fobjects_to_push\x18\x03 \x03(\t\x1a.\n\x0cResultsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\"\x1e\n\tReconnect\x12\x11\n\treconnect\x18\x01 \x01(\x04\x32\xa0\x08\n\x05\x46leet\x12]\n\x0cRegisterNode\x12$.flwr.proto.RegisterNodeFleetRequest\x1a%.flwr.proto.RegisterNodeFleetResponse\"\x00\x12S\n\x0c\x41\x63tivateNode\x12\x1f.flwr.proto.ActivateNodeRequest\x1a .flwr.proto.ActivateNodeResponse\"\x00\x12Y\n\x0e\x44\x65\x61\x63tivateNode\x12!.flwr.proto.DeactivateNodeRequest\x1a\".flwr.proto.DeactivateNodeResponse\"\x00\x12\x63\n\x0eUnregisterNode\x12&.flwr.proto.UnregisterNodeFleetRequest\x1a\'.flwr.proto.UnregisterNodeFleetResponse\"\x00\x12\x62\n\x11SendNodeHeartbeat\x12$.flwr.proto.SendNodeHeartbeatRequest\x1a%.flwr.proto.SendNodeHeartbeatResponse\"\x00\x12S\n\x0cPullMessages\x12\x1f.flwr.proto.PullMessagesRequest\x1a .flwr.proto.PullMessagesResponse\"\x00\x12S\n\x0cPushMessages\x12\x1f.flwr.proto.PushMessagesRequest\x1a .flwr.proto.PushMessagesResponse\"\x00\x12\x41\n\x06GetRun\x12\x19.flwr.proto.GetRunRequest\x1a\x1a.flwr.proto.GetRunResponse\"\x00\x12\x41\n\x06GetFab\x12\x19.flwr.proto.GetFabRequest\x1a\x1a.flwr.proto.GetFabResponse\"\x00\x12M\n\nPushObject\x12\x1d.flwr.proto.PushObjectRequest\x1a\x1e.flwr.proto.PushObjectResponse\"\x00\x12M\n\nPullObject\x12\x1d.flwr.proto.PullObjectRequest\x1a\x1e.flwr.proto.PullObjectResponse\"\x00\x12q\n\x16\x43onfirmMessageReceived\x12).flwr.proto.ConfirmMessageReceivedRequest\x1a*.flwr.proto.ConfirmMessageReceivedResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'flwr.proto.fleet_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_PUSHMESSAGESRESPONSE_RESULTSENTRY']._loaded_options = None
  _globals['_PUSHMESSAGESRESPONSE_RESULTSENTRY']._serialized_options = b'8\001'
  _globals['_REGISTERNODEFLEETREQUEST']._serialized_start=159
  _globals['_REGISTERNODEFLEETREQUEST']._serialized_end=205
  _globals['_REGISTERNODEFLEETRESPONSE']._serialized_start=207
  _globals['_REGISTERNODEFLEETRESPONSE']._serialized_end=251
  _globals['_ACTIVATENODEREQUEST']._serialized_start=253
  _globals['_ACTIVATENODEREQUEST']._serialized_end=322
  _globals['_ACTIVATENODERESPONSE']._serialized_start=324
  _globals['_ACTIVATENODERESPONSE']._serialized_end=363
  _globals['_DEACTIVATENODEREQUEST']._serialized_start=365
  _globals['_DEACTIVATENODEREQUEST']._serialized_end=405
  _globals['_DEACTIVATENODERESPONSE']._serialized_start=407
  _globals['_DEACTIVATENODERESPONSE']._serialized_end=431
  _globals['_UNREGISTERNODEFLEETREQUEST']._serialized_start=433
  _globals['_UNREGISTERNODEFLEETREQUEST']._serialized_end=478
  _globals['_UNREGISTERNODEFLEETRESPONSE']._serialized_start=480
  _globals['_UNREGISTERNODEFLEETRESPONSE']._serialized_end=509
  _globals['_PULLMESSAGESREQUEST']._serialized_start=511
  _globals['_PULLMESSAGESREQUEST']._serialized_end=585
  _globals['_PULLMESSAGESRESPONSE']._serialized_start=588
  _globals['_PULLMESSAGESRESPONSE']._serialized_end=750
  _globals['_PUSHMESSAGESREQUEST']._serialized_start=753
  _globals['_PUSHMESSAGESREQUEST']._serialized_end=904
  _globals['_PUSHMESSAGESRESPONSE']._serialized_start=907
  _globals['_PUSHMESSAGESRESPONSE']._serialized_end=1108
  _globals['_PUSHMESSAGESRESPONSE_RESULTSENTRY']._serialized_start=1062
  _globals['_PUSHMESSAGESRESPONSE_RESULTSENTRY']._serialized_end=1108
  _globals['_RECONNECT']._serialized_start=1110
  _globals['_RECONNECT']._serialized_end=1140
  _globals['_FLEET']._serialized_start=1143
  _globals['_FLEET']._serialized_end=2199
# @@protoc_insertion_point(module_scope)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface `build` command."""


import hashlib
import zipfile
from io import BytesIO
from pathlib import Path
from typing import Annotated, Any

import pathspec
import tomli
import tomli_w
import typer

from flwr.common.constant import (
    FAB_CONFIG_FILE,
    FAB_DATE,
    FAB_EXCLUDE_PATTERNS,
    FAB_HASH_TRUNCATION,
    FAB_INCLUDE_PATTERNS,
    FAB_MAX_SIZE,
)

from .config_utils import load_and_validate
from .utils import build_pathspec, is_valid_project_name, load_gitignore_patterns


def write_to_zip(
    zipfile_obj: zipfile.ZipFile, filename: str, contents: bytes | str
) -> zipfile.ZipFile:
    """Set a fixed date and write contents to a zip file.

    Parameters
    ----------
    zipfile_obj : zipfile.ZipFile
        The ZipFile object to write to.
    filename : str
        Name of the file within the zip archive.
    contents : bytes | str
        The file contents to write.

    Returns
    -------
    ZipFile
        The modified ZipFile object.
    """
    zip_info = zipfile.ZipInfo(filename)
    zip_info.date_time = FAB_DATE
    zipfile_obj.writestr(zip_info, contents)
    return zipfile_obj


def get_fab_filename(config: dict[str, Any], fab_hash: str) -> str:
    """Get the FAB filename based on the given config and FAB hash.

    Parameters
    ----------
    config : dict[str, Any]
        The project configuration dictionary.
    fab_hash : str
        The SHA-256 hash of the FAB file.

    Returns
    -------
    str
        The formatted FAB filename in the pattern:
        <publisher>.<name>.<version>.<hash_prefix>.fab
    """
    publisher = config["tool"]["flwr"]["app"]["publisher"]
    name = config["project"]["name"]
    version = config["project"]["version"].replace(".", "-")
    fab_hash_truncated = fab_hash[:FAB_HASH_TRUNCATION]
    return f"{publisher}.{name}.{version}.{fab_hash_truncated}.fab"


# pylint: disable=too-many-locals, too-many-statements
def build(
    app: Annotated[
        Path | None,
        typer.Option(help="Path of the Flower App to bundle into a FAB"),
    ] = None,
) -> None:
    """Build a Flower App into a Flower App Bundle (FAB).

    You can run ``flwr build`` without any arguments to bundle the app located in the
    current directory. Alternatively, you can you can specify a path using the ``--app``
    option to bundle an app located at the provided path. For example:

    ``flwr build --app ./apps/flower-hello-world``.
    """
    if app is None:
        app = Path.cwd()

    app = app.resolve()
    if not app.is_dir():
        typer.secho(
            f"❌ The path {app} is not a valid path to a Flower app.",
            fg=typer.colors.RED,
            bold=True,
            err=True,
        )
        raise typer.Exit(code=1)

    if not is_valid_project_name(app.name):
        typer.secho(
            f"❌ The project name {app.name} is invalid, "
            "a valid project name must start with a letter, "
            "and can only contain letters, digits, and hyphens.",
            fg=typer.colors.RED,
            bold=True,
            err=True,
        )
        raise typer.Exit(code=1)

    config, errors, warnings = load_and_validate(app / "pyproject.toml")
    if config is None:
        typer.secho(
            "Project configuration could not be loaded.\npyproject.toml is invalid:\n"
            + "\n".join([f"- {line}" for line in errors]),
            fg=typer.colors.RED,
            bold=True,
            err=True,
        )
        raise typer.Exit(code=1)

    if warnings:
        typer.secho(
            "Project configuration is missing the following "
            "recommended properties:\n" + "\n".join([f"- {line}" for line in warnings]),
            fg=typer.colors.RED,
            bold=True,
        )

    # Build FAB
    fab_bytes = build_fab_from_disk(app)

    # Calculate hash for filename
    fab_hash = hashlib.sha256(fab_bytes).hexdigest()

    # Get the name of the zip file
    fab_filename = get_fab_filename(config, fab_hash)

    # Write the FAB
    Path(fab_filename).write_bytes(fab_bytes)

    typer.secho(
        f"🎊 Successfully built {fab_filename}", fg=typer.colors.GREEN, bold=True
    )


def build_fab_from_disk(app: Path) -> bytes:
    """Build a FAB from files on disk and return the FAB as bytes.

    This function reads files from disk and bundles them into a FAB.

    Parameters
    ----------
    app : Path
        Path to the Flower app to bundle into a FAB.

    Returns
    -------
    bytes
        The FAB as bytes.
    """
    app = app.resolve()

    # Collect all files recursively (including pyproject.toml and .gitignore)
    all_files = [f for f in app.rglob("*") if f.is_file()]

    # Create dict mapping relative paths to Path objects
    files_dict: dict[str, bytes | Path] = {
        # Ensure consistent path separators across platforms
        str(file_path.relative_to(app)).replace("\\", "/"): file_path
        for file_path in all_files
    }

    # Build FAB from the files dict
    return build_fab_from_files(files_dict)


def build_fab_from_files(files: dict[str, bytes | Path]) -> bytes:
    r"""Build a FAB from in-memory files and return the FAB as bytes.

    This is the core FAB building function that works with in-memory data.
    It accepts either bytes or Path objects as file contents, applies filtering
    rules (include/exclude patterns), and builds the FAB.

    Parameters
    ----------
    files : dict[str, Union[bytes, Path]]
        Dictionary mapping relative file paths to their contents.
        - Keys: Relative paths (strings)
        - Values: Either bytes (file contents) or Path (will be read)
        Must include "pyproject.toml" and optionally ".gitignore".

    Returns
    -------
    bytes
        The FAB as bytes.

    Examples
    --------
    Build a FAB from in-memory files::

        files = {
            "pyproject.toml": b"[project]\nname = 'myapp'\n...",
            ".gitignore": b"*.pyc\n__pycache__/\n",
            "src/client.py": Path("/path/to/client.py"),
            "src/server.py": b"print('hello')",
            "README.md": b"# My App\n",
        }
        fab_bytes = build_fab_from_files(files)
    """

    def to_bytes(content: bytes | Path) -> bytes:
        return content.read_bytes() if isinstance(content, Path) else content

    # Extract, load, and parse pyproject.toml
    if FAB_CONFIG_FILE not in files:
        raise ValueError(f"{FAB_CONFIG_FILE} not found in files")
    pyproject_content = to_bytes(files[FAB_CONFIG_FILE])
    config = tomli.loads(pyproject_content.decode("utf-8"))

    # Remove the 'federations' field if it exists
    if (
        "tool" in config
        and "flwr" in config["tool"]
        and "federations" in config["tool"]["flwr"]
    ):
        del config["tool"]["flwr"]["federations"]

    # Extract and load .gitignore if present
    gitignore_content = None
    if ".gitignore" in files:
        gitignore_content = to_bytes(files[".gitignore"])

    # Get exclude and include specs
    exclude_spec = get_fab_exclude_pathspec(gitignore_content)
    include_spec = get_fab_include_pathspec()

    # Filter files based on include/exclude specs
    filtered_paths = [
        path.replace("\\", "/")  # Ensure consistent path separators across platforms
        for path in files.keys()
        if include_spec.match_file(path) and not exclude_spec.match_file(path)
    ]
    filtered_paths.sort()  # Sort for deterministic output

    # Create a zip file in memory
    list_file_content = ""

    fab_buffer = BytesIO()
    with zipfile.ZipFile(fab_buffer, "w", zipfile.ZIP_DEFLATED) as fab_file:
        # Add pyproject.toml
        write_to_zip(fab_file, FAB_CONFIG_FILE, tomli_w.dumps(config))

        for file_path in filtered_paths:

            # Get file contents as bytes
            file_content = to_bytes(files[file_path])

            # Write file to FAB
            write_to_zip(fab_file, file_path, file_content)

            # Calculate file info for CONTENT manifest
            sha256_hash = hashlib.sha256(file_content).hexdigest()
            file_size_bits = len(file_content) * 8  # size in bits
            list_file_content += f"{file_path},{sha256_hash},{file_size_bits}\n"

        # Add CONTENT manifest to the zip file
        write_to_zip(fab_file, ".info/CONTENT", list_file_content)

    fab_bytes = fab_buffer.getvalue()

    # Validate FAB size
    if len(fab_bytes) > FAB_MAX_SIZE:
        raise ValueError(
            f"FAB size exceeds maximum allowed size of {FAB_MAX_SIZE:,} bytes. "
            "To reduce the package size, consider ignoring unnecessary files "
            "via your `.gitignore` file or excluding them from the build."
        )

    return fab_bytes


def get_fab_include_pathspec() -> pathspec.PathSpec:
    """Get the PathSpec for files to include in a FAB.

    Returns
    -------
    PathSpec
        PathSpec object with default include patterns for FAB files.
    """
    return build_pathspec(FAB_INCLUDE_PATTERNS)


def get_fab_exclude_pathspec(gitignore_content: bytes | None) -> pathspec.PathSpec:
    """Get the PathSpec for files to exclude from a FAB.

    If gitignore_content is provided, its patterns will be combined with the default
    exclude patterns.

    Parameters
    ----------
    gitignore_content : bytes | None
        Optional gitignore file content as bytes.

    Returns
    -------
    PathSpec
        PathSpec object with combined exclude patterns.
    """
    patterns = list(FAB_EXCLUDE_PATTERNS)
    if gitignore_content:
        patterns += load_gitignore_patterns(gitignore_content)
    return pathspec.PathSpec.from_lines("gitwildmatch", patterns)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Ed25519-only asymmetric cryptography utilities."""

import base64
from pathlib import Path

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import ed25519


def generate_key_pair() -> tuple[ed25519.Ed25519PrivateKey, ed25519.Ed25519PublicKey]:
    """Generate an Ed25519 private/public key pair.

    Returns
    -------
    Tuple[Ed25519PrivateKey, Ed25519PublicKey]
        Private and public key pair.
    """
    private_key = ed25519.Ed25519PrivateKey.generate()
    return private_key, private_key.public_key()


def private_key_to_bytes(private_key: ed25519.Ed25519PrivateKey) -> bytes:
    """Serialize an Ed25519 private key to PEM bytes.

    Parameters
    ----------
    private_key : Ed25519PrivateKey
        The private key to serialize.

    Returns
    -------
    bytes
        PEM-encoded private key.
    """
    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


def bytes_to_private_key(private_key_bytes: bytes) -> ed25519.Ed25519PrivateKey:
    """Deserialize an Ed25519 private key from PEM bytes.

    Parameters
    ----------
    private_key_bytes : bytes
        PEM-encoded private key.

    Returns
    -------
    Ed25519PrivateKey
        Deserialized private key.
    """
    return serialization.load_pem_private_key(
        private_key_bytes, password=None
    )  # type: ignore[return-value]


def public_key_to_bytes(public_key: ed25519.Ed25519PublicKey) -> bytes:
    """Serialize an Ed25519 public key to PEM bytes.

    Parameters
    ----------
    public_key : Ed25519PublicKey
        The public key to serialize.

    Returns
    -------
    bytes
        PEM-encoded public key.
    """
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )


def bytes_to_public_key(public_key_bytes: bytes) -> ed25519.Ed25519PublicKey:
    """Deserialize an Ed25519 public key from PEM bytes.

    Parameters
    ----------
    public_key_bytes : bytes
        PEM-encoded public key.

    Returns
    -------
    Ed25519PublicKey
        Deserialized public key.
    """
    return serialization.load_pem_public_key(public_key_bytes)  # type: ignore[return-value]


def sign_message(private_key: ed25519.Ed25519PrivateKey, message: bytes) -> bytes:
    """Sign a message using an Ed25519 private key.

    Parameters
    ----------
    private_key : Ed25519PrivateKey
        The private key used for signing.
    message : bytes
        The message to sign.

    Returns
    -------
    bytes
        The signature of the message.
    """
    return private_key.sign(message)


def verify_signature(
    public_key: ed25519.Ed25519PublicKey, message: bytes, signature: bytes
) -> bool:
    """Verify a signature using an Ed25519 public key.

    Parameters
    ----------
    public_key : Ed25519PublicKey
        The public key used for verification.
    message : bytes
        The original message.
    signature : bytes
        The signature to verify.

    Returns
    -------
    bool
        True if the signature is valid, False otherwise.
    """
    try:
        public_key.verify(signature, message)
        return True
    except InvalidSignature:
        return False


def create_message_to_sign(fab_digest: bytes, timestamp: int) -> bytes:
    """Create a canonical message:
    timestamp (8 bytes big-endian) + fab_digest.
    """
    timestamp_bytes = timestamp.to_bytes(8, byteorder="big")
    return timestamp_bytes + fab_digest


def decode_base64url(sig: str) -> bytes:
    """Convert signature to b64 format."""
    # add missing padding (=) to a multiple of 4
    pad = (-len(sig)) % 4
    return base64.urlsafe_b64decode(sig + ("=" * pad))


def load_private_key(path: Path) -> ed25519.Ed25519PrivateKey:
    """Load an SSH-format private key (Ed25519) using cryptography."""
    key_bytes = path.read_bytes()
    private_key = serialization.load_ssh_private_key(key_bytes, password=None)
    if not isinstance(private_key, ed25519.Ed25519PrivateKey):
        raise ValueError("Private key is not Ed25519")
    return private_key



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""GrpcAdapter implementation."""


import signal
import time
from logging import DEBUG
from typing import Any, TypeVar, cast

import grpc
from google.protobuf.message import Message as GrpcMessage

from flwr.common import log
from flwr.common.constant import (
    GRPC_ADAPTER_METADATA_FLOWER_PACKAGE_NAME_KEY,
    GRPC_ADAPTER_METADATA_FLOWER_PACKAGE_VERSION_KEY,
    GRPC_ADAPTER_METADATA_FLOWER_VERSION_KEY,
    GRPC_ADAPTER_METADATA_MESSAGE_MODULE_KEY,
    GRPC_ADAPTER_METADATA_MESSAGE_QUALNAME_KEY,
    GRPC_ADAPTER_METADATA_SHOULD_EXIT_KEY,
)
from flwr.common.version import package_name, package_version
from flwr.proto.fab_pb2 import GetFabRequest, GetFabResponse  # pylint: disable=E0611
from flwr.proto.fleet_pb2 import (  # pylint: disable=E0611
    ActivateNodeRequest,
    ActivateNodeResponse,
    DeactivateNodeRequest,
    DeactivateNodeResponse,
    PullMessagesRequest,
    PullMessagesResponse,
    PushMessagesRequest,
    PushMessagesResponse,
    RegisterNodeFleetRequest,
    RegisterNodeFleetResponse,
    UnregisterNodeFleetRequest,
    UnregisterNodeFleetResponse,
)
from flwr.proto.grpcadapter_pb2 import MessageContainer  # pylint: disable=E0611
from flwr.proto.grpcadapter_pb2_grpc import GrpcAdapterStub
from flwr.proto.heartbeat_pb2 import (  # pylint: disable=E0611
    SendNodeHeartbeatRequest,
    SendNodeHeartbeatResponse,
)
from flwr.proto.message_pb2 import (  # pylint: disable=E0611
    ConfirmMessageReceivedRequest,
    ConfirmMessageReceivedResponse,
    PullObjectRequest,
    PullObjectResponse,
    PushObjectRequest,
    PushObjectResponse,
)
from flwr.proto.run_pb2 import GetRunRequest, GetRunResponse  # pylint: disable=E0611
from flwr.supercore.constant import FORCE_EXIT_TIMEOUT_SECONDS

T = TypeVar("T", bound=GrpcMessage)


class GrpcAdapter:
    """Adapter class to send and receive gRPC messages via the ``GrpcAdapterStub``.

    This class utilizes the ``GrpcAdapterStub`` to send and receive gRPC messages
    which are defined and used by the Fleet API, as defined in ``fleet.proto``.
    """

    def __init__(self, channel: grpc.Channel) -> None:
        self.stub = GrpcAdapterStub(channel)

    def _send_and_receive(
        self, request: GrpcMessage, response_type: type[T], **kwargs: Any
    ) -> T:
        # Serialize request
        req_cls = request.__class__
        container_req = MessageContainer(
            metadata={
                GRPC_ADAPTER_METADATA_FLOWER_PACKAGE_NAME_KEY: package_name,
                GRPC_ADAPTER_METADATA_FLOWER_PACKAGE_VERSION_KEY: package_version,
                GRPC_ADAPTER_METADATA_FLOWER_VERSION_KEY: package_version,
                GRPC_ADAPTER_METADATA_MESSAGE_MODULE_KEY: req_cls.__module__,
                GRPC_ADAPTER_METADATA_MESSAGE_QUALNAME_KEY: req_cls.__qualname__,
            },
            grpc_message_name=req_cls.__qualname__,
            grpc_message_content=request.SerializeToString(),
        )

        # Send via the stub
        container_res = cast(
            MessageContainer, self.stub.SendReceive(container_req, **kwargs)
        )

        # Handle control message
        should_exit = (
            container_res.metadata.get(GRPC_ADAPTER_METADATA_SHOULD_EXIT_KEY, "false")
            == "true"
        )
        if should_exit:
            log(
                DEBUG,
                'Received shutdown signal: exit flag is set to ``"true"``. Exiting...',
            )
            signal.raise_signal(signal.SIGTERM)
            # Give some time to handle the signal
            time.sleep(FORCE_EXIT_TIMEOUT_SECONDS + 1)

        # Check the grpc_message_name of the response
        if container_res.grpc_message_name != response_type.__qualname__:
            raise ValueError(
                f"Invalid grpc_message_name. Expected {response_type.__qualname__}"
                f", but got {container_res.grpc_message_name}."
            )

        # Deserialize response
        response = response_type()
        response.ParseFromString(container_res.grpc_message_content)
        return response

    def RegisterNode(  # pylint: disable=C0103
        self, request: RegisterNodeFleetRequest, **kwargs: Any
    ) -> RegisterNodeFleetResponse:
        """."""
        return self._send_and_receive(request, RegisterNodeFleetResponse, **kwargs)

    def ActivateNode(  # pylint: disable=C0103
        self, request: ActivateNodeRequest, **kwargs: Any
    ) -> ActivateNodeResponse:
        """."""
        return self._send_and_receive(request, ActivateNodeResponse, **kwargs)

    def DeactivateNode(  # pylint: disable=C0103
        self, request: DeactivateNodeRequest, **kwargs: Any
    ) -> DeactivateNodeResponse:
        """."""
        return self._send_and_receive(request, DeactivateNodeResponse, **kwargs)

    def UnregisterNode(  # pylint: disable=C0103
        self, request: UnregisterNodeFleetRequest, **kwargs: Any
    ) -> UnregisterNodeFleetResponse:
        """."""
        return self._send_and_receive(request, UnregisterNodeFleetResponse, **kwargs)

    def SendNodeHeartbeat(  # pylint: disable=C0103
        self, request: SendNodeHeartbeatRequest, **kwargs: Any
    ) -> SendNodeHeartbeatResponse:
        """."""
        return self._send_and_receive(request, SendNodeHeartbeatResponse, **kwargs)

    def PullMessages(  # pylint: disable=C0103
        self, request: PullMessagesRequest, **kwargs: Any
    ) -> PullMessagesResponse:
        """."""
        return self._send_and_receive(request, PullMessagesResponse, **kwargs)

    def PushMessages(  # pylint: disable=C0103
        self, request: PushMessagesRequest, **kwargs: Any
    ) -> PushMessagesResponse:
        """."""
        return self._send_and_receive(request, PushMessagesResponse, **kwargs)

    def GetRun(  # pylint: disable=C0103
        self, request: GetRunRequest, **kwargs: Any
    ) -> GetRunResponse:
        """."""
        return self._send_and_receive(request, GetRunResponse, **kwargs)

    def GetFab(  # pylint: disable=C0103
        self, request: GetFabRequest, **kwargs: Any
    ) -> GetFabResponse:
        """."""
        return self._send_and_receive(request, GetFabResponse, **kwargs)

    def PushObject(  # pylint: disable=C0103
        self, request: PushObjectRequest, **kwargs: Any
    ) -> PushObjectResponse:
        """."""
        return self._send_and_receive(request, PushObjectResponse, **kwargs)

    def PullObject(  # pylint: disable=C0103
        self, request: PullObjectRequest, **kwargs: Any
    ) -> PullObjectResponse:
        """."""
        return self._send_and_receive(request, PullObjectResponse, **kwargs)

    def ConfirmMessageReceived(  # pylint: disable=C0103
        self, request: ConfirmMessageReceivedRequest, **kwargs: Any
    ) -> ConfirmMessageReceivedResponse:
        """."""
        return self._send_and_receive(request, ConfirmMessageReceivedResponse, **kwargs)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower message-based FedXgbCyclic strategy."""


from collections.abc import Callable, Iterable
from logging import INFO
from typing import cast

from flwr.common import (
    ArrayRecord,
    ConfigRecord,
    Message,
    MessageType,
    MetricRecord,
    RecordDict,
    log,
)
from flwr.server import Grid

from .fedavg import FedAvg
from .strategy_utils import sample_nodes


# pylint: disable=line-too-long
class FedXgbCyclic(FedAvg):
    """Configurable FedXgbCyclic strategy implementation.

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        fraction_train: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_available_nodes: int = 2,
        weighted_by_key: str = "num-examples",
        arrayrecord_key: str = "arrays",
        configrecord_key: str = "config",
        train_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        evaluate_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
    ) -> None:
        super().__init__(
            fraction_train=fraction_train,
            fraction_evaluate=fraction_evaluate,
            min_train_nodes=2,
            min_evaluate_nodes=2,
            min_available_nodes=min_available_nodes,
            weighted_by_key=weighted_by_key,
            arrayrecord_key=arrayrecord_key,
            configrecord_key=configrecord_key,
            train_metrics_aggr_fn=train_metrics_aggr_fn,
            evaluate_metrics_aggr_fn=evaluate_metrics_aggr_fn,
        )

        self.registered_nodes: dict[int, int] = {}

        if fraction_train not in (0.0, 1.0):
            raise ValueError(
                "fraction_train can only be set to 1.0 or 0.0 for FedXgbCyclic."
            )
        if fraction_evaluate not in (0.0, 1.0):
            raise ValueError(
                "fraction_evaluate can only be set to 1.0 or 0.0 for FedXgbCyclic."
            )

    def _reorder_nodes(self, node_ids: list[int]) -> list[int]:
        """Re-order node ids based on registered nodes.

        Each node ID is assigned a persistent index in `self.registered_nodes`
        the first time it appears. The input list is then reordered according
        to these stored indices, and the result is compacted into ascending
        order (1..N) for the current call.
        """
        # Assign new indices to unknown nodes
        next_index = max(self.registered_nodes.values(), default=0) + 1
        for nid in node_ids:
            if nid not in self.registered_nodes:
                self.registered_nodes[nid] = next_index
                next_index += 1

        # Sort node_ids by their stored indices
        sorted_by_index = sorted(node_ids, key=lambda x: self.registered_nodes[x])

        # Compact re-map of indices just for this output list
        unique_indices = sorted(self.registered_nodes[nid] for nid in sorted_by_index)
        remap = {old: new for new, old in enumerate(unique_indices, start=1)}

        # Build the result list ordered by compact indices
        result_list = [
            nid
            for _, nid in sorted(
                (remap[self.registered_nodes[nid]], nid) for nid in sorted_by_index
            )
        ]
        return result_list

    def _make_sampling(
        self, grid: Grid, server_round: int, configure_type: str
    ) -> list[int]:
        """Sample nodes using the Grid."""
        # Sample nodes
        num_nodes = int(len(list(grid.get_node_ids())) * self.fraction_train)
        sample_size = max(num_nodes, self.min_train_nodes)
        node_ids, _ = sample_nodes(grid, self.min_available_nodes, sample_size)

        # Re-order node_ids
        node_ids = self._reorder_nodes(node_ids)

        # Sample the clients sequentially given server_round
        sampled_idx = (server_round - 1) % len(node_ids)
        sampled_node_id = [node_ids[sampled_idx]]

        log(
            INFO,
            f"{configure_type}: Sampled %s nodes (out of %s)",
            len(sampled_node_id),
            len(node_ids),
        )
        return sampled_node_id

    def configure_train(
        self, server_round: int, arrays: ArrayRecord, config: ConfigRecord, grid: Grid
    ) -> Iterable[Message]:
        """Configure the next round of federated training."""
        # Sample one node
        sampled_node_id = self._make_sampling(grid, server_round, "configure_train")

        # Always inject current server round
        config["server-round"] = server_round

        # Construct messages
        record = RecordDict(
            {self.arrayrecord_key: arrays, self.configrecord_key: config}
        )
        return self._construct_messages(record, sampled_node_id, MessageType.TRAIN)

    def aggregate_train(
        self,
        server_round: int,
        replies: Iterable[Message],
    ) -> tuple[ArrayRecord | None, MetricRecord | None]:
        """Aggregate ArrayRecords and MetricRecords in the received Messages."""
        valid_replies, _ = self._check_and_log_replies(replies, is_train=True)

        arrays, metrics = None, None
        if valid_replies:
            reply_contents = [msg.content for msg in valid_replies]
            array_record_key = next(iter(reply_contents[0].array_records.keys()))

            # Fetch the client model from current round as global model
            arrays = cast(ArrayRecord, reply_contents[0][array_record_key])

            # Aggregate MetricRecords
            metrics = self.train_metrics_aggr_fn(
                reply_contents,
                self.weighted_by_key,
            )
        return arrays, metrics

    def configure_evaluate(
        self, server_round: int, arrays: ArrayRecord, config: ConfigRecord, grid: Grid
    ) -> Iterable[Message]:
        """Configure the next round of federated evaluation."""
        # Sample one node
        sampled_node_id = self._make_sampling(grid, server_round, "configure_evaluate")

        # Always inject current server round
        config["server-round"] = server_round

        # Construct messages
        record = RecordDict(
            {self.arrayrecord_key: arrays, self.configrecord_key: config}
        )
        return self._construct_messages(record, sampled_node_id, MessageType.EVALUATE)
